#!/usr/local/bin/perl5 -w
#
# Bias.pm -- Class to extract satellite and receiver biases.
#

package Bias;
use strict;

# Implemented by referencing GFO bias information.
#   
# Use:
#   (private) read   -- to read in biases for a specific date. 
#   (private) exist  -- to test existence of a bias (satellite or receiver).
#   extract -- extract a specific bias (satellite or receiver).
#   cache  -- cache biases in associative array (satellite or receiver).
#   print  -- to print out bias (satellite or receiver).
#
# Assumptions:
#   Satellite ID is a integer under 100.
#   Receiver ID is a 4 character string.
#
# Note: Function overloading is not supported in PERL 5.004 (see Class::Multimethods at CPAN).
#       Methods exist, extract, cache and print must use extra logic to overcome this omission.
# See Bias_test.pl for example usage.

sub new {
  my $this = shift;
  my $arg = shift;
  my $file = "/iono/gfo/uarc/biasesk81hdbentYYMMDD.dat";	# Use only if implicit filename is supplied.
  
  if ($arg =~ /^[0-9]{6}/) { 
    $file =~ s/YYMMDD/$arg/;
    die "Bias file not found for date: $arg." unless ( -e $file);
  } else {
    $file = $arg;
    die "Bias file not found: $file." unless ( -e $file);
  }

  my $arrhash = [{}, {}];		# SATs and RCVs
  $this->read($file, $arrhash);		# Initialize

  return bless $arrhash;
}
  
sub read {
  my($this, $file, $arrhash) = @_;

  open (BIAS, "<$file");
  while(<BIAS>) {
    my ($name, $bias) = (split(' ', $_))[0,2];
    $name =~ m/([A-Z0-9]{4})([A-Z0-9]{2})/ || die "error in reading bias file.";
    my $h = $1;
    my $t = $2;
    ${$arrhash->[0]}{$t} = $bias if ($h =~ m/SATL/);	# SATs 	
    ${$arrhash->[1]}{$h} = $bias if ($t =~ m/XX/);	# RCVs
  }
  close(BIAS);
}

sub exist {
  my ($this, $name) = @_;
  my $exist = 1;
  # eval will not trap warnings in $@, use signal handler instead.
  local $SIG{__WARN__} = sub { $exist = 0; };
  local $SIG{__DIE__} = sub { $exist = 0; };

  if (length $name < 4 ) {
    eval "${$this->[0]}{$name}"; 		# SATs
  } else {
    eval "${$this->[1]}{$name}"; 		# RCVs
  }
  
  $exist;
}

sub extract {
  my ($this, $name) = @_;
  my $bias;

  return "" if ! $this->exist($name);

  if (length $name < 4 ) {
    $bias = ${$this->[0]}{$name};	# SATs
  } else {
    $bias = ${$this->[1]}{$name};	# RCVs
  }
  $bias;
}

sub print {
  my ($this, $name) = @_;

  return print "Bias not found for $name.\n" if ! $this->exist($name);

  if (length $name < 4 ) {
    print "Bias for satellite " . $name . ": " . ${$this->[0]}{$name} . "\n";
  } else {
    print "Bias for receiver " . $name . ": ". ${$this->[1]}{$name} . "\n";
  }
}

sub cache {
  my ($this, $type) = @_;
  
  if ($type =~ m/^SAT/) {
    return $this->[0];
  } elsif ($type =~ m/^RCV/) {
    return $this->[1];
  } else {
    die "Incorrect argument to cache method: $type";
  }
}

1;
