#!/usr/local/bin/perl -w
#
# DateTime.pm -- Date & Time class based on Timetrans library.
#                Represents a Date/Time as seconds past J2000 epoch (see JDREF constant)
#                and provides various format conversions and date delta arithmetic.
#

package DateTime;
use strict;
use Carp;

# Implemented as many small methods following Timetrans library.
# Object state is saved as a reference to a double value containing
# seconds past epoch.
#
# See DateTime_test.pl for examples of usage.
#

use constant JDREF  => 2451545.0;          # Julian date of J2000 reference epoch
use constant SECDAY => 86400.0;
use constant FRAC_FORMAT => "%8.6f";

my @MONTHS     = ( 'JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN',
                   'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC' );
my %MONTH_STRS = ( 'JAN' => '01', 'FEB' => '02', 'MAR' => '03',
                   'APR' => '04', 'MAY' => '05', 'JUN' => '06',
                   'JUL' => '07', 'AUG' => '08', 'SEP' => '09',
                   'OCT' => '10', 'NOV' => '11', 'DEC' => '12' );

my @INTERVALS = ( 'decade', 'year', 'mon', 'day', 'hour', 'min', 'sec', 'frac' );

sub new {    # construct DateTime object from another DateTime object,
             #                        or from (int or double) seconds past epoch
    my($class, $init) = @_;
    my($init_sec);

    if ( defined $init ) {
	if ( ref $init ) { $init_sec = $$init }    # arg is another DateTime object
        if ( scalar $init ) { $init_sec = $init }  # arg is value of seconds past epoch
    } else {
        $init_sec = 0.0;                           # no arg
    }
    return bless( \$init_sec, ref($class) || $class );
}

sub fromCal {    # init DateTime object from 7 inputs below
    my($self, $year, $mon, $day, $hour, $min, $sec, $frac) = @_;

    $$self = calsec($year, $mon, $day, $hour, $min, $sec, $frac);
    return $self;
}

sub toCal {    # convert DateTime to array of 7 values below
    my($self) = @_;
    my($year, $mon, $day, $hour, $min, $sec, $frac);
    my($s_, $jd, $tmp);

    $s_ = $$self;          # extract object state into $s_ variable
    return seccal($s_);
}

sub fromCalStr {    # init DateTime object from Calendar String yyyy/mm/dd,hh:mm:ss.frac
    my($self, $str) = @_;
    my($year, $mon, $day, $hour, $min, $sec, $frac);

    if ( $str =~
          m#([0-9]{2,4})/([0-9]{2})/([0-9]{2}).?([0-9]{0,2}):?([0-9]{0,2}):?([0-9]{0,2})(\.?[0-9]*)# ) {
        ($year, $mon, $day, $hour, $min, $sec, $frac) = ($1, $2, $3, $4, $5, $6, $7);
	if ( $hour eq "" ) { $hour = 0 }
	if ( $min  eq "" ) { $min  = 0 }
	if ( $sec  eq "" ) { $sec  = 0 }
        if ( $frac eq "" ) { $frac = 0.0 }
        if ( $year > 99 && $year < 1900 ) {
	    croak "DateTime: Bad year field, $year, in CalStr => $str";
            return;
	} else {
            return $self->fromCal($year, $mon, $day, $hour, $min, $sec, $frac);
	}
     } else {
	 croak "DateTime: Bad CalStr => $str";
	 return;
     }
}

sub toCalStr {    # convert DateTime to standard Calendar string yyyy/mm/dd,hh:mm:ss.frac
    my($self) = @_;
    my($str, $frac_str);

    my ($year, $mon, $day, $hour, $min, $sec, $frac) = $self->toCal();
    $frac_str = sprintf FRAC_FORMAT, $frac;
    $str = sprintf "%4.4d/%2.2d/%2.2d,%2.2d:%2.2d:%2.2d", 
                   $year, $mon, $day, $hour, $min, $sec, $frac;
    $str .= substr($frac_str, 1) if $frac != 0;
    return $str;
}

sub fromShortDate { # init DateTime object from short date format yymmdd
    my($self, $yymmdd) = @_;
    my $cal_str = year_4digit(substr($yymmdd, 0, 2)) . "/" . 
                      substr($yymmdd, 2, 2) . "/" .
                      substr($yymmdd, 4, 2);
    return (new DateTime)->fromCalStr($cal_str);
}

sub toShortDate { # convert DateTime object to short date format yymmdd
    my($self) = @_;
    my $cal_str = $self->toCalStr();
    return substr($cal_str,2,2) . substr($cal_str,5,2) .
           substr($cal_str,8,2);
}

sub toHoursInDay {
    my($self) = @_;
    
    my ($year, $mon, $day, $hour, $min, $sec, $frac) = $self->toCal();
    my $hours = $hour + $min/60. + ($sec + $frac)/3600.;
    return sprintf "%10.7f", $hours;
}

sub toSecondsInDay {
    my($self) = @_;
    
    my ($year, $mon, $day, $hour, $min, $sec, $frac) = $self->toCal();
    my $seconds = $hour*60.*60. + $min*60. + $sec + $frac;
    return sprintf "%11.1f", $seconds;
}

sub toSecPastTime {  # Returns delta seconds since a specified date/time.
    my($self, $time2) = @_;
    my $sec2 = new DateTime->fromCalStr($time2)->toSec();
    return $$self - $sec2;
}

sub toHoursPastTime {  # Returns hours (float) since a specified date/time.
    my($self, $time2) = @_;
    my $delta_sec = $self->toSecPastTime($time2);
    my ($sign, $day, $hour, $min, $sec, $frac) = dsec2delta($delta_sec);
    my $hours = 24*$day + $hour + $min/60. + ($sec + $frac)/3600.;
    $hours = -$hours if $sign eq '-';
    return $hours;
}

sub toDaysPastTime {  # Returns days (float) since a specified date/time.
    my($self, $time2) = @_;
    my $hours = $self->toHoursPastTime($time2);
    return $hours / 24.;
}

sub DOY {
    my($self) = @_;
    my($year, $mon, $day, $hour, $min, $sec, $frac) = $self->toCal();
    return ymd2doy($year, $mon, $day);
}

sub fromCh {    # init DateTime object from Char String dd-mon-yyyy hh:mm:ss.frac
    my($self, $str) = @_;
    my($year, $mon, $day, $hour, $min, $sec, $frac);

    ($year, $mon, $day, $hour, $min, $sec, $frac) = ch2cal($str);
    $$self = calsec($year, $mon, $day, $hour, $min, $sec, $frac);
    return $self;
}

sub toCh {    # convert DateTime to standard Char String dd-mon-yyyy hh:mm:ss.frac
    my($self) = @_;
    my($year, $mon, $day, $hour, $min, $sec, $frac);
    my($str, $frac_str, $mon_str);

    ($year, $mon, $day, $hour, $min, $sec, $frac) = $self->toCal();
    $mon_str = $MONTHS[$mon-1];
    $frac_str = sprintf FRAC_FORMAT, $frac;
    $str = sprintf "%2.2d-%s-%4.4d %2.2d:%2.2d:%2.2d", 
                   $day, $mon_str, $year, $hour, $min, $sec, $frac;
    $str .= substr($frac_str, 1);
    return $str;
}

sub fromSec {    # init DateTime from seconds past epoch
    my($self, $sec) = @_;

    $$self = $sec;
    return $self;
}

sub toSec {    # convert DateTime to seconds past epoch
    my($self) = @_;
    return $$self;
}

sub floor {  # Round date/time down to specified interval (year, mon, day, hour, min, sec).
    my($self, $interval) = @_;

    my($year, $mon, $day, $hour, $min, $sec, $frac) = seccal($$self);
    if (      $interval =~ /year/i ) {
        $self->fromCal($year, 0, 0, 0, 0, 0, 0.0);
    } elsif ( $interval =~ /mon/i  ) {
        $self->fromCal($year, $mon, 0, 0, 0, 0, 0.0);
    } elsif ( $interval =~ /day/i  ) {
        $self->fromCal($year, $mon, $day, 0, 0, 0, 0.0);
    } elsif ( $interval =~ /hour/i ) {
        $self->fromCal($year, $mon, $day, $hour, 0, 0, 0.0);
    } elsif ( $interval =~ /min/i  ) {
        $self->fromCal($year, $mon, $day, $hour, $min, 0, 0.0);
    } elsif ( $interval =~ /sec/i  ) {
        $self->fromCal($year, $mon, $day, $hour, $min, $sec, 0.0);
    } else {
	die "DateTime::floor: Bad interval, $interval (year|mon|day|hour|min|sec)";
    }
    return $self;
}

sub addDeltaTime {    # add Delta Time [+-]dddd/hh:mm:ss.frac to the DateTime
    my($self, $str) = @_;
    my($sec);

    $sec = delta2dsec($str);
    $self->addDeltaSec($sec);
}

sub addDeltaSec {    # add delta of (+ or -) seconds to internal seconds past epoch
    my($self, $delta_sec) = @_;

    $$self += $delta_sec;
    return $self;
}

sub calsec {    # convert 7 inputs below to seconds past epoch
    my($year, $mon, $day, $hour, $min, $sec, $frac) = @_;
    my($jd, $s);

    $year = year_4digit($year);
    $jd = date2j($year, $mon, $day) - 0.5;
    $s = jd2sec($jd) + $hour*3600. + $min*60. + $sec + $frac;
    return $s;
}

sub year_4digit {
    my($year) = @_;

    if ($year < 100 ) {
        $year += ( $year <= 68 ? 2000 : 1900);
    }
    return $year;
}

sub seccal {    # convert seconds past epoch to array of 7 values
    my($s) = @_;
    my($year, $mon, $day, $hour, $min, $sec, $frac);
    my($jd, $tmp);

    $frac = $s - int($s);
    if ( $frac < 0.0 ) { $frac += 1.0 }
    $jd = sec2jd( $s - $frac + 0.5 );
    ($year, $mon, $day, $hour, $min, $sec, $tmp) =  jd2cal($jd);

    return ($year, $mon, $day, $hour, $min, $sec, $frac);
}

sub sec2jd {    # convert seconds past epoch to Julian date
    my($s) = @_;
    my($jd);
    $jd = $s / SECDAY + JDREF;
    return $jd;
}

sub jd2cal {    # convert Julian date to array of 7 Cal values
    my($jd) = @_;
    my($year, $mon, $day, $hour, $min, $sec, $frac);
    my($jdplus, $jdint, $dsec, $isec);

    $jdplus = $jd + 0.5;
    $jdint = int($jdplus);
    $dsec = SECDAY * ($jdplus - int($jdplus));
    if ( $dsec >= SECDAY ) { 
	$jdint += 1;
	$dsec -= SECDAY;
    }
    $isec = int($dsec);
    ($year, $mon, $day) = j2date($jdint);
    $hour = int($isec / 3600);
    $isec -= $hour*3600;
    $min = int($isec/60);
    $isec -= $min*60;
    $sec = $isec;
    $frac = $dsec - int($dsec);
    return ($year, $mon, $day, $hour, $min, $sec, $frac);
}

sub date2j {    # convert YMD date to integer Julian date (internal)
    my($year, $mon, $day) = @_;
    my($jd, $c, $ya);

    if ( $mon > 2 ) { 
	$mon -= 3;
    } else {
	$mon += 9;
	$year -= 1;
    }
    $c = int($year/100);
    $ya = $year - $c*100;
    $jd = int((146097*$c)/4) + int((1461*$ya)/4) + int((153*$mon+2)/5) + $day + 1721119;
    return $jd;
}

sub j2date {    # convert integer Julian date to calendar date as (y,m,d)
    my($jd) = @_;
    my($j, $y, $m, $d);

    $j = int($jd);
    $j -= 1721119;
    $y = int( (4*$j - 1) / 146097 );
    $j = 4*$j - 1 - 146097*$y;
    $d = int( $j / 4 );
    $j = int( (4*$d + 3) / 1461 );
    $d = 4*$d+ 3 - 1461*$j;
    $d = int( ($d+4) / 4 );
    $m = int( (5*$d - 3) / 153 );
    $d = 5*$d - 3 - 153*$m;
    $d = int( ($d+5) / 5 );
    $y = 100*$y + $j;
    if ( $m < 10 ) {
	$m += 3;
    } else {
	$m -= 9;
	$y += 1;
    }
    return ($y, $m, $d);
}

sub jd2sec {    # convert double Julian date to double seconds past epoch
    my($jd) = @_;
    my($s);

    $s = ($jd - JDREF) * SECDAY;
    return $s;
}

sub ch2cal {    # convert Char String dd-mon-yyyy hh:mm:ss.frac to 7 Cal values
    my($str) = @_;
    my($year, $mon, $day, $hour, $min, $sec, $frac, $mon_str, $n);

    if ( $str =~
          m#([0-9]{1,2})\-([a-zA-Z]{3})\-([0-9]{4}).?([0-9]{0,2}):?([0-9]{0,2}):?([0-9]{0,2})(\.?[0-9]*)# ) {
        ($day, $mon_str, $year, $hour, $min, $sec, $frac) = ($1, $2, $3, $4, $5, $6, $7);
        $mon_str =~ tr/a-z/A-Z/;
        if ( defined( ($mon = $MONTH_STRS{$mon_str}) ) ) {
            if ( $hour eq "" ) { $hour = 0 }
            if ( $min  eq "" ) { $min  = 0 }
            if ( $sec  eq "" ) { $sec  = 0 }
            if ( $frac eq "" ) { $frac = 0.0 }
            return ($year, $mon, $day, $hour, $min, $sec, $frac);
	} else {
	    croak "DateTime: Bad Month in ChStr => $str";
	}
     } else {
	 croak "DateTime: Bad ChStr => $str";
     }
}

sub ch2sec {    # convert Char String dd-mon-yyyy hh:mm:ss.frac to seconds past epoch
    my($str) = @_;

    my($year, $mon, $day, $hour, $min, $sec, $frac) = ch2cal($str);
    return calsec($year, $mon, $day, $hour, $min, $sec, $frac);
}

sub delta2dsec {    # convert Delta Time [+-]dddd/hh:mm:ss.frac to delta seconds
    my($str) = @_;
    my ($ds, $sign, $day, $hour, $min, $sec, $frac, $sign_str);

    $sign = 1;
    if ( $str =~
          m#([-+]?)([0-9]{0,4})/?([0-9]{0,2}):?([0-9]{0,2}):?([0-9]{0,2})(\.?[0-9]*)# ) {
        ($sign_str, $day, $hour, $min, $sec, $frac) = ($1, $2, $3, $4, $5, $6);
        if ( $sign_str eq "-" ) { $sign = -1 }
        if ( $day  eq "" ) { $day  = 0 }
        if ( $hour eq "" ) { $hour = 0 }
        if ( $min  eq "" ) { $min  = 0 }
        if ( $sec  eq "" ) { $sec  = 0 }
        if ( $frac eq "" ) { $frac = 0.0 }
        $ds = $day*86400. + $hour*3600. +  $min*60. + $sec + $frac;
        $ds *= $sign;
    } else {
	croak "DateTime: Bad Delta Time string => $str";
    }
}

sub dsec2delta {    # convert delta seconds to 6 Delta Time Values
                   # ($sign, $day, $hour, $min, $sec, $frac)
    my($ds) = @_;
    my ($sign, $day, $hour, $min, $sec, $frac);

    $sign = $ds >= 0. ? "+" : "-";
    ($day, $ds) = mod_remainder($ds, 86400);
    ($hour, $ds) = mod_remainder($ds, 3600);
    ($min, $ds) = mod_remainder($ds, 60);
    ($sec, $ds) = mod_remainder($ds, 1);
    $frac = $ds;
    return ($sign, $day, $hour, $min, $sec, $frac);
}

sub dsec2delta_str {    # convert delta seconds to Delta Time [+-]dddd/hh:mm:ss.frac
    my($ds) = @_;
    my ($frac_str, $str);

    my ($sign, $day, $hour, $min, $sec, $frac) = sec2delta($ds);
    $frac_str = sprintf FRAC_FORMAT, $frac;
    $str = sprintf "%s%d/%2.2d:%2.2d:%2.2d", $sign, $day, $hour, $min, $sec;
    $str .= substr($frac_str, 1);
}

sub dsec2hours {
    my($ds) = @_;
    my ($sign, $day, $hour, $min, $sec, $frac) = dsec2delta($ds);
    my $hours = 24*$day + $hour + $min/60. + ($sec + $frac)/3600.;
    $hours = -$hours if $sign eq '-';
    return $hours;
}

sub dsec2days {
    my($ds) = @_;
    my $hours = dsec2hours($ds);
    return $hours / 24.;
}

sub mod_remainder {    # extract an integer multiple of modulo base and return integer & remainder
    my($s, $m) = @_;
    my($n, $t);

    $n = int( $s / $m );
    $t = $s - $n * $m;
    return ($n, $t);
}

sub ymd2doy {  # Compute integer day of year from year, month, and day integers.
    my($year, $mon, $day) = @_;
    $year = year_4digit($year);
    my $doy = date2j($year, $mon, $day) - date2j($year, 1, 0);
    return sprintf("%3.3d", $doy);
}

sub lesser_interval {
    my($interval) = @_;
    my $i = interval_index($interval);
    return $INTERVALS[$i+1];
}

sub greater_interval {
    my($interval) = @_;
    my $i = interval_index($interval);
    return $INTERVALS[$i-1];
}

sub interval_index {
    my($interval) = @_;
    for (my $i=0; $i <= $#INTERVALS; $i++) {
	return $i if $interval =~ /$INTERVALS[$i]/i;
    }
    croak "DateTime::interval_index: Bad interval name, $interval";
}

1;


