#!/usr/local/bin/perl -w
#
# DatedValues.pm -- Keep timetagged data values as ASCII lines of text where
#                   the timetag is the first column and one or more data values
#                   follow in additional columns.  The file is sorted using an
#                   input keylen so one can use leading data columns as additional
#                   sort keys (e.g., time, stn, sat).
#
#                   Provides find and insert/update routines.
#

package DatedValues;
use strict;
use DateTime;
use Carp;

my %MODE_KEYLENS = (
    date   => 10,   # YYYY/MM/DD
    hour   => 13,   # YYYY/MM/DD,HH
    minute => 16,   # YYYY/MM/DD,HH:MM
    second => 19,   # YYYY/MM/DD,HH:MM:SS
    time   => 19,   # YYYY/MM/DD,HH:MM:SS
);

sub new {  # Create DatedValues object from a filename.
    my($class, $file, $mode, $keylen) = @_;
    
    croak "DatedValues: Bad or missing file: $file"
	unless defined $file && -e $file;
    croak "DatedValues: Bad Date/Time mode: $mode"
        unless defined $mode && defined $MODE_KEYLENS{$mode};
    $keylen = $MODE_KEYLENS{$mode} if !defined $keylen;

    my %hash = (file => $file, mode => $mode, keylen => $keylen);
    return bless( \%hash, ref($class) || $class );
}


sub find {  # Find a timetagged line by key (date/time and possible additional keys).
            # Can be multiple results if mult_mode is set.
    my($self, $key, $mult_mode) = @_;
    my $file = $$self{'file'};

    $key = $self->fix_time($key);
    my $lines = `grep -i '$key' $file`;
    chomp $lines;
    if ( ! defined $mult_mode ) {
        croak "DatedValues::find: Multiple matches found, expected one: $lines"
            if index($lines, "\n") >= 0;
    }
    return $lines;
}


sub find_range {  # Find the range of timetagged lines between two times (inclusive).
                  # Additional key fields can be passed as part of the two keys for
                  # files that contain multiple records for each time epoch.
    my($self, $key1, $key2) = @_;
    my $file = $$self{'file'};
    my $keylen = $$self{'keylen'};

    $key1 = substr($self->fix_time($key1), 0, $keylen);
    $key2 = substr($self->fix_time($key2), 0, $keylen);
    my $lines = "";
    open(IN, "$file") || croak "DatedValues::find_range: Cannot open file: $file";

    $_ = <IN>;
    while ( defined $_ && (substr($_,0,$keylen) lt $key1 || $_ =~ /^#/) ) {
        $_ = <IN>;
    }
    while ( defined $_ && substr($_,0,$keylen) le $key2 ) {
	$lines .= $_;
        $_ = <IN>;
    }
    close(IN);
    chomp $lines;
    return $lines;
}


sub insert {
    my($self, $lines, $replace_mode) = @_;
    my $file = $$self{'file'};
    my $mode = $$self{'mode'};
    my $keylen = $$self{'keylen'};

    my $key;

    if (!$replace_mode) { $replace_mode = "replace"; }

    `cp -p $file.bak $file.bak2`;
    `cp -p $file $file.bak`;

    ##Read in existing file into hash
    my %master;

    open(IN, "$file.bak") || croak "DatedValues::insert: Cannot open backup file: $file.bak";
    foreach my $incoming (<IN>) { $master{substr($incoming,0,$keylen)}=$incoming; }
    close(IN);

    ##Split out incoming lines to add into an array
    my (@lines, $line);
    if ( index($lines, "\n") > 0 ) {
	@lines = split "\n", $lines;
    } else {
	$lines[0] = $lines;
    }

    ##Add new data array to hash
    foreach $line (@lines)
    {
     $line = $self->fix_time($line);
     $key = substr($line,0,$keylen);
     if ($master{$key} && (lc($replace_mode) ne "replace"))
     {croak "DatedValues::insert Error, record already exists:\n=>$line\nin file: $file"}
     else { $master{$key}=$line."\n"; }
    } 
 
    ##Write hash back out
    open(OUT, "> $file") || croak "DatedValues::insert: Cannot open file: $file";
    foreach $line (sort keys %master)
     { 
	 print OUT $master{$line};
     }    
    close(OUT);
}

sub fix_time {  # Truncate or extend date/time key to match format in file.
    my($self, $key) = @_;
    my $mode = $$self{'mode'};

    my $time = $key;
    my $rest;
    my $i = index($key, ' ');
    if ( $i >= 0 ) {
        $time = substr($key, 0, $i);
        $rest = substr($key, $i);
    }
    my $dt = new DateTime->fromCalStr($time);
    $time = $dt->toCalStr();
    $key = substr($time, 0, $MODE_KEYLENS{$mode});
    $key .= $rest if defined $rest;
    return $key;
}


sub append_alt_time {  # Append alternative time tag (float) to each line for plotting.
                       # Time Type = day (default), hour, minute, second.
                       # Time Mode = absolute (default) or relative.
    my($lines, $time_type, $time_mode) = @_;
    $time_type = "day" if !defined $time_type;
    $time_mode = "abs" if !defined $time_mode;

    chomp $lines if $lines =~ /\n$/;
    my @lines = split "\n", $lines;
    $lines = "";
    my $first = 1;
    my $dt = new DateTime();
    my ($time, $rest, $i, $dsec, $sec, $sec0, $year, $greater_time_type, $alt_time);
    foreach my $line ( @lines ) {
        $i = index($line, ' ');
        $time = substr($line, 0, $i);
        $rest = substr($line, $i);
	if ( $first ) {
	    $first = 0;
	    if ( $time_mode =~ /abs/i ) {
		$greater_time_type = DateTime::greater_interval($time_type);
                $sec0 = $dt->fromCalStr($time)->floor($greater_time_type)->toSec();
	    } elsif ( $time_mode =~ /rel/i ) {
                $sec0 = $dt->fromCalStr($time)->floor($time_type)->toSec();
#		$year = substr($time, 0, 4);
#		$sec0 = $dt->fromCalStr("$year/01/01,00:00:00")->toSec();
	    } else {
		croak "DatedValues::append_alt_time: Bad time mode, $time_mode (abs|rel)";
	    }
	}

        $sec = $dt->fromCalStr($time)->toSec();
	$dsec = $sec - $sec0;
	if ( $time_type =~ /day/i ) {
	    $alt_time = DateTime::dsec2days($dsec);
	} elsif ( $time_type =~ /hour/i ) {
	    $alt_time = DateTime::dsec2hours($dsec);
	} elsif ( $time_type =~ /min/i ) {
	    $alt_time = DateTime::dsec2hours($dsec) * 60.;
	} elsif ( $time_type =~ /sec/i ) {
            $alt_time = $dsec;
	} else {
            croak "DatedValues::append_alt_time: Bad time type, $time_type (day|hour|min|sec)";
	}
	$lines .= $line . ' ' . $alt_time . "\n";
    }
    chomp $lines;
    return $lines;
}

1;

