#!/usr/local/bin/perl5 -w
#
# Db.pm -- Database wrapper class:  hides DBI complexities.
#

package Db;
use strict;
use Carp;

# Save Db access state variables in a hash:
#   host     = host where Db resides
#   db       = name of database
#   driver   = Db driver type, e.g., "mysql"
#   user     = username on Db, not necessarily related to login name
#   password = password for that user, grants access privileges to selected databases
#
#   table       = name of table to access
#   field_names = @array of names (strings), 
#   field_types = @array of types (strings)

# Use:
#   new      -- must specify host, db, driver, user, & password
#   connect  -- CONNECT to Db using set host, db, driver, user, & password
#   create   -- CREATE a table automatically using table, field_names, & field_types
#   create_custom -- specify a custom SQL CREATE cmd
#   prep_insert  -- prepare an SQL INSERT cmd with placeholders
#   insert -- insert a record into the table
#   prep_select -- prepare a SELECT cmd with parameters
#   fetch -- execute SELECT cmd to fetch a row of table values
#   set_field_names -- specify  FIELD names for created table
#   set_field_types -- specify  FIELD types for created table
#   get_field_names -- retrieve FIELD names from existing table
#   get_field_types -- retreive FIELD types from existing table
#

sub new {
    $class = shift;
    my $self = {};
    bless $self, $class;
    $self->init( @_ );
    return $self;
}

sub init {
    croak "Db: Must specify host, db, driver, user, password, & table" if @_ < 5;
    $self->{ qw(host db driver user password table) } = @_;
    $self->{"connected"} = 0;
    $self->{"have_fields"} = 0;
}

sub connect {
    
    croak "Db: Missing arg among host, db, driver, user, password, & table"
        if $self->{ qw(host db driver user password table) } < 6;
    my $dsn = "DBI:$self->{"driver"}:database=$self->{"db"};host=$self->{"host"}";
    my $dbh = DBI->connect($dsn, $self->{"user"}, $self->{"password"}, 
                               { RaiseError => 1} );
    $self->{"dbh"} = $dbh;
    return $dbh;
}

1;
