#!/usr/local/bin/perl -w
#
# GeophyIndices.pm -- Geophysical Indices class to hold routines for retrieving
#                     and manipulating geomagnetic and solar indices such as
#                     F10.7, Ap, Kp, etc.
#                     Retrieves prompt and archival index values from NOAA's SEC.
#
# Added predict ability 02-23-04 Mandrake

package GeophyIndices;
use strict;
#use LWP;
use LWP::Simple;
use DatedValues;
use DateTime;
use Carp;

# Web sites for prompt geomagnetic and solar data at SEC.
use constant SEC_SWx_SITE => 'http://sec.noaa.gov/ftpdir/indices/';
use constant SEC_DGD   => 'DGD.txt';       # daily geomagnetic data
use constant SEC_DSD   => 'DSD.txt';       # daily solar data

# Local cache of F10.7 (& SSN) values for computing 45-day means.
my $F10p7_SSN_FILE = $ENV{'F10p7_SSN_FILE'};
my $AP_FILE = $ENV{'AP_FILE'};
my $DST_FILE = $ENV{'DST_FILE'};
my $DAYS_IN_F10p7_MEAN = 30;

sub get_prompt_Ap_Kp { # Get daily planetary Ap & 3-hour Kp for date YYYY/MM/DD
    my ($date, $rKp) = @_;

    croak "get_prompt_Ap_Kp: Bad date input: $date"
        unless $date =~ m#[0-9]{4}/[0-9]{2}/[0-9]{2}#;
    (my $sdate = $date) =~ s#/# #g;         # sdate now contains YYYY MM DD

    # Find Ap & Kp indices in the DGD file for the specified date.
    my $url = SEC_SWx_SITE . SEC_DGD;
#    my $agent = LWP::UserAgent->new;
#    my $request = HTTP::Request->new( GET => $url );
#    my $response = $agent->request($request);
#    print STDERR $response->code, "\n";
#    print STDERR $response->content, "\n";
#    $response->is_success or die "get_prompt_Kp_Ap: $url: ", $response->message, "\n";
#    my @lines = split( "\n", $response->content);
    my $content = get($url);
    defined $content or die "get_prompt_Kp_Ap: Bad remote GET: $url\n";
    my @lines = split( "\n", $content);

    my $found_Ap = 0;
    my $found_hdr = 0;
    my $year = substr( $date, 0, 4);
    my $Ap = "-1";
    @$rKp = qw(-1 -1 -1 -1 -1 -1 -1 -1);
    for ( my $i = 0; $i <= $#lines; $i++ ) {
        $_ = $lines[$i];
        if ( !$found_hdr && /^$year/ ) {
	    for ( my $j = -3; $j < 0; $j++ ) { print STDERR $lines[$i + $j], "\n" }
            $found_hdr++;
        }
        if ( /^$sdate/ ) {
            print $lines[$i], "\n";
	    my $str = substr($lines[$i], 59, 21);
            ($Ap, @$rKp) = unpack "A3xA2A2A2A2A2A2A2A2", $str;
            $found_Ap++ if $Ap != -1;
        }
    }

    print STDERR "get_prompt_Ap_Kp: Ap = $Ap  on $date\n";
    print STDERR join " ", "get_prompt_Ap_Kp: Kp (3-hour) = ", @$rKp, " on $date\n";
    $found_Ap or carp "get_prompt_Ap_Kp: Geomag indices not found for this date: $date";
    return $Ap;
}


sub get_prompt_F10p7_SSN { # Get daily solar F10.7 flux and sunspot number for date YYYY/MM/DD
    my ($date) = @_;

    croak "get_prompt_F10p7_SSN: Bad date input: $date"
        unless $date =~ m#[0-9]{4}/[0-9]{2}/[0-9]{2}#;
    (my $sdate = $date) =~ s#/# #g;         # sdate now contains YYYY MM DD

    # Find F10.7 & SSN indices in the DSD file for the specified date.
    my $url = SEC_SWx_SITE . SEC_DSD;
    my $content = get($url);
    defined $content or die "get_prompt_F10p7_SSN: Bad remote GET: $url\n";
    my @lines = split( "\n", $content);

    my $found_F10p7 = 0;
    my $found_hdr = 0;
    my $year = substr( $date, 0, 4);
    my $F10p7 = "-1";
    my $SSN = "-1";
    for ( my $i = 0; $i <= $#lines; $i++ ) {
        $_ = $lines[$i];
        if ( !$found_hdr && /^$year/ ) {
	    for ( my $j = -4; $j < 0; $j++ ) { print STDERR $lines[$i + $j], "\n" }
            $found_hdr++;
        }
        if ( /^$sdate/ ) {
            print $lines[$i], "\n";
	    $F10p7 = substr($lines[$i], 12, 3);
	    $SSN   = substr($lines[$i], 19, 3);
            $found_F10p7++ if $F10p7 != -1;
        }
    }

    print STDERR "get_prompt_F10p7_SSN: F10.7 = $F10p7  on $date\n";
    print STDERR "get_prompt_F10p7_SSN: SSN   = $SSN  on $date\n";
    $found_F10p7 or carp "get_prompt_F10p7_SSN: Geomag indices not found for this date: $date";
    return ($F10p7, $SSN);
}


sub save_F10_SSN_value {  # Save F10p7 & SSN values for one day.
    my($date, $F10p7, $SSN) = @_;
    my $dv = new DatedValues($F10p7_SSN_FILE, "date");
    my $line = $date . ' ' . $F10p7 . ' ' . $SSN;
    $dv->insert($line, "replace");
}

sub update_F10_Ap_predict { # Gather F10.7 and Ap predict values from NOAA's 45 day forecast
    my ($F10_file, $Ap_file) = @_;
    $F10_file = $ENV{'F10p7_SSN_FILE'} if !defined $F10_file;
    $Ap_file  = $ENV{'AP_FILE'} if !defined $Ap_file;
   
    # Get F10.7 & Ap 45 day forecast
    my $url = "http://www.sec.noaa.gov/ftpdir/latest/45DF.txt";
    my $content = get($url);
    defined $content or die "update_F10_Ap_predict: Bad remote GET: $url\n";
    my @lines = split ("\n",$content);
    
    #Set up the Ap dated values object
    my $dv = new DatedValues($Ap_file, "date");
    my $dt = new DateTime;
    my @store;

    foreach my $line (@lines)
    {
     if ($line =~ /45-DAY F10.7 CM FLUX FORECAST/ ) 
       {
	 $dv->insert(join("\n",@store),"replace"); #Add all gathered values at once
         $dv = new DatedValues($F10_file,"date");
         @store=();
	 print "switch\n";
       } #switch to reading F10.7 values

     if ($line =~ /^\d\d\D\D\D\d\d/ ) 
      { #a data line
       my @pairs = split(' ',$line);
       for (my $t=0;$t<=$#pairs;$t+=2)
       { 
  	   my $date = $pairs[$t  ];
           my $val  = $pairs[$t+1];
           $date=~/(\d\d)(\D\D\D)(\d\d)/;
           $date=substr($dt->fromCh($1.'-'.$2.'-20'.$3)->toCalStr(),0,10);
           print "date: $date\n";
	   push @store, $date.' '.$val.' 0 Predict'; # Add the appropriate value to the file, including a zero for possible SSN and P for predict
       }
      }
    }
    $dv->insert(join("\n",@store),"replace");
}

sub update_F10_SSN_values_from_prompt {  # Retrieve and store latest 30 daily F10.7 values.
    my($file) = @_;
    $file = $ENV{'F10p7_SSN_FILE'} if !defined $file;

    # Get F10.7 & SSN indices in the DSD file.
    my $url = SEC_SWx_SITE . SEC_DSD;
    my $content = get($url);
    defined $content or die "update_F10p7_SSN_values_from_prompt: Bad remote GET: $url\n";
    my @lines = split( "\n", $content);

    my ($date, $F10p7, $SSN, @vals, $vals);
    while ( defined ($_ = shift @lines) ) {
	if ( /^([0-9]{4} [0-9]{2} [0-9]{2}) +([0-9]+) +([0-9]+)/ ) {
	    $date = $1;  $F10p7 = $2;  $SSN = $3;
            $date =~ s# #/#g;
            $F10p7 = sprintf "%3d", $F10p7;
            $SSN = sprintf "%3d", $SSN;
	    push @vals, $date . ' '  . $F10p7 . ' ' . $SSN;
	}
    }

    $vals = join "\n", @vals;
    my $dv = new DatedValues($file, "date");

    print "changing: $vals\n";
    $dv->insert($vals, "replace");
}


sub update_Ap_values_from_prompt {  # Retrieve and store latest 30 daily Ap values.
    my($file) = @_;
    $file = $ENV{'$AP_FILE'} if !defined $file;

    # Get F10.7 & SSN indices in the DSD file.
    my $url = SEC_SWx_SITE . SEC_DGD;
    my $content = get($url);
    defined $content or die "update_Ap_values_from_prompt: Bad remote GET: $url\n";
    my @lines = split( "\n", $content);

    my ($date, $Ap, @vals, $vals);
    while ( defined ($_ = shift @lines) ) {
	if ( /^([0-9]{4} [0-9]{2} [0-9]{2}) +([0-9]+) +([0-9]+)/ ) {
	    $date = $1;
            $date =~ s# #/#g;
            $Ap = substr($_, 59, 3);
	    push @vals, $date . ' '  . $Ap;
	}
    }

    $vals = join "\n", @vals;
    my $dv = new DatedValues($file, "date");
    $dv->insert($vals, "replace");
}


sub calc_F10p7_mean {  # Compute F10.7 running mean over n days ending at last date
    my($last_date, $n) = @_;
    $n = $DAYS_IN_F10p7_MEAN if !defined $n;

    my $dt = new DateTime->fromCalStr($last_date)->addDeltaTime( -$n+1 );
    my $first_date = substr($dt->toCalStr(), 0, 10);
    my $dv = new DatedValues($F10p7_SSN_FILE, "date");
    my $lines = $dv->find_range($first_date, $last_date);
    my @lines = split "\n", $lines;
    my $sum = 0.;
    my $i = 0;
    my ($date, $F10p7, $SSN);
    foreach my $line ( @lines ) {
	($date, $F10p7, $SSN) = split ' ', $line;
	$sum += $F10p7;
	$i++;
    }
    if ( $n != $i ) {
	croak "GeophyIndices::calc_F01p7_mean: Cannot find $n F10.7 values to average ($i)";
    }
    return sprintf "%5.1f", $sum / $i;
}


sub update_Dst_from_archive {  # Update accumulating Dst file from NGDC yearly archival file.
    my($in_file, $out_file) = @_;
    $out_file = $DST_FILE if !defined $out_file;

    open(IN, $in_file) || die "GeophyIndices::update_Dst_from_archive: Bad input file, $in_file";
    my ($yr, $mon, $day, $hour, $base, $vals, @vals);
    my $dt = new DateTime;
    my $lines = "";
    while ( <IN> ) {
	if ( /^DST([0-9]{2})([0-9]{2}).([0-9]{2})/ ) {
	    $yr = $1;  $mon = $2;  $day = $3;
	    $base = substr($_, 16, 4);   # base value to be added to following hourly values
	    $vals = substr($_, 20);
	    @vals = unpack "A4" x 24, $vals;    #hourly Dst values
            for ($hour=0; $hour < 24; $hour++ ) {
	        $vals[$hour] += $base;
                $lines .= $dt->fromCal($yr, $mon, $day, $hour, 30, 0, 0.0)->toCalStr();
		# Assign time of 00:30 to Dst value for 0 to 1 hour, etc.
		$lines .= ' ' . sprintf("%4d", $vals[$hour]) . "\n"
            }
	}
    }

    my $dv = new DatedValues($out_file, "minute");
    chomp $lines if $lines =~ /\n$/;
    $dv->insert($lines, "replace");
}

sub pick_drift_pattern {
    my $cal_str = shift @_;
    my $dt = (new DateTime)->fromCalStr($cal_str);
    my ($yr, $mon, $day, $hr, $min, $sec, $frac) = $dt->toCal();
    
    # Solar levels (min, mod, or max) for years 1990 to 2014.
    my @SOLAR_LEVELS = qw(max max max mod mod min min min mod mod mod max max max
                          mod mod min min min mod mod mod max max max);
    my $solar = $SOLAR_LEVELS[$yr - 1990];

    # Seasonal drift pattern for each month.
    my @MONTHLY_DRIFTS = qw(dec mar mar mar jun jun jun mar mar mar dec dec);
    my $month = $MONTHLY_DRIFTS[$mon-1];
    my $pattern = "drift_am_" . $month . "." . $solar;
    return $pattern;
}

1;
