#!/usr/local/bin/perl -w
#
# Rinex2.pm -- Small library to read/write RINEX obs & nav files.
#             Extracts key fields from the header (see read_obs_header) and
#             can be used to extract a time series of data fields for each
#             GPS satellite.
#

package Rinex2;

use strict;
use DateTime;
use Connect;
use Carp;
use constant DEBUG => 1;

# Implemented as a package of functions, such as read_obs_header, read_obs_block,
# merge_rinex_obs, read_nav_header, read_nav_block, etc.
#
# See Rinex_test.pl for examples of usage.

use constant NIL      => "Nil";             # blank obs fields set to NIL string
use constant C        => 2.99792458e8;      # speed of light in m/s
use constant F1R      => 154;               # L1 frequency ratio
use constant F2R      => 120;               # L2 frequency ratio
use constant CHIPRATE => 10.23e6;           # GPS chip rate in Hz
use constant F1       => F1R * CHIPRATE;    # L1 frequency in c/s
use constant F2       => F2R * CHIPRATE;    # L2 frequency in c/s
use constant W1       => C / F1;            # L1 wavelength in m
use constant W2       => C / F2;            # L2 wavelength in m
use constant SEC2TECU => 2.853340e9;        # seconds to TECU conversion

use constant Y_JUMP_HUGE => 1.e32;            # huge jump threshold for disabling connection
use constant Y_JUMP_L1   => 1.e6;             # jump threshold for connecting L1/L2 phase
use constant Y_JUMP_LI   =>   20;             # jump threshold for connecting LI phase
#use constant Y_JUMP_LC   =>   20;             # jump threshold for connecting LC phase

sub read_obs_header {    # Read the header of a RINEX OBS file from a file handle ($fh).
    my($fh, $h) = @_;    # Important fields are returned in the hash passed by reference ($h).
    my($key, @f);

    $h->{'ALL'} = "";
    while ( ($_ = <$fh>) !~ m/END OF HEADER/i ) {
	$_ =~ s/\r\n$/\n/;            # if have DOS CR/LF line endings, delete CR
	$h->{'ALL'} .= $_;            # accumulate header lines in ALL field
        $key = substr($_, 60);
	chomp($key);  chomp($_);
        if ( $key !~ m/COMMENT/ ) {
            $h->{$key} = $_;          # save line under LABEL at end of line, except for COMMENTs
	}
    }
    $h->{'ALL'} .= $_;                # save END OF HEADER line

    # extract some important fields and save values in hash (under lowercase keys)
    if ( defined $h->{'RINEX VERSION / TYPE'} ) {
        ($h->{'version'}, $h->{'type'}) 
            = split ' ', substr($h->{'RINEX VERSION / TYPE'}, 0, 60);
    }
    if ( defined $h->{'PGM / RUN BY / DATE'} ) {
        ($h->{'pgm_name'}, $h->{'pgm_date'}, $h->{'run_by'}, $h->{'run_date'}, $h->{'run_time'})
	    =  split ' ', substr($h->{'PGM / RUN BY / DATE'}, 0, 60);
    }
    if ( defined $h->{'# / TYPES OF OBSERV'} ) {
        @f = split ' ', substr($h->{'# / TYPES OF OBSERV'}, 0, 60);
	my $n_types = shift @f;
	$h->{'n_obs_types'} = $n_types;
        if ( @f != $n_types ) {
	    croak "Rinex::read_obs_header: Invalid '# / TYPES OF OBSERV' line.";
	}
        $h->{'obs_types'} = join " ", @f;
    } else {
	croak "Rinex::read_obs_header: Incomplete RINEX header, must have '# / TYPES OF OBSERV' line.";
    }
    if ( defined $h->{'TIME OF FIRST OBS'} ) {
        $h->{'first_time'} = substr($h->{'TIME OF FIRST OBS'}, 0, 60);
    }
    if ( defined $h->{'TIME OF LAST OBS'} ) {
        $h->{'last_time'} = substr($h->{'TIME OF LAST OBS'}, 0, 60);
    }
    if ( defined $h->{'REC # / TYPE / VERS'} ) {
        @f = split ' ', substr($h->{'REC # / TYPE / VERS'}, 0, 60);
        $h->{'rcvr_no'} = shift @f;
        $h->{'rcvr_type'} = join " ", splice(@f, 0, 2);
        $h->{'rcvr_version'} = join " ", @f;
    }
    if ( defined $h->{'MARKER NAME'} ) {
        ($h->{'site'}) = split ' ', substr($h->{'MARKER NAME'}, 0, 60);
        if ( $h->{'site'} eq "" ) {
	    croak "Rinex::read_obs_header: Invalid or missing MARKER NAME (site).";
        }
    } else {
	croak "Rinex::read_obs_header: Incomplete RINEX header, must have 'MARKER NAME' line.";
    }
    if ( defined $h->{'MARKER NUMBER'} ) {
        ($h->{'site_no'}) = split ' ', substr($h->{'MARKER NUMBER'}, 0, 60);
    }

    if ( DEBUG ) {
        print STDERR "\nRinex Header info:\n";
        foreach $key (sort keys %$h) {
	    if ( $key =~ /[a-z]+/ ) {
	        print STDERR $key, " = ", $h->{$key}, "\n";
            }
        }
        print STDERR "\n";
    }
}


sub update_obs_header {    # update times of first and last obs in the header lines
    my($hdr_in, $begin_obs_time_str, $end_obs_time_str) = @_;

    # append header lines up to "TIME OF .* OBS" line or "END OF HEADER" line
    my $hdr_out = "";
    my @line = split "\n", $hdr_in;
    $_ = shift @line;
    while ( $_  &&  ! m/TIME OF .* OBS/  &&  ! m/END OF HEADER/ ) {
        $hdr_out .= $_ . "\n";
        $_ = shift @line;
    }
    unshift @line, $_;

    # insert TIME OF <FIRST|LAST> OBS lines where they usually are in the header, or at end
    my $begin_hdr_time_str = obs_time_str2hdr_time_str($begin_obs_time_str);
    my $end_hdr_time_str = obs_time_str2hdr_time_str($end_obs_time_str);
    $hdr_out .= $begin_hdr_time_str . " " x 5 . "GPS" . " " x 9 . "TIME OF FIRST OBS\n";
    $hdr_out .= $end_hdr_time_str . " " x 5 . "GPS" . " " x 9 . "TIME OF LAST OBS\n";

    # append rest of header
    while ( ($_ = shift @line) ) {
        $hdr_out .= $_ . "\n" if ! m/TIME OF .* OBS/;
    }
    return $hdr_out;
}


sub read_obs_blocks {                          # Read sat obs blocks of a RINEX OBS file from a file handle ($fh).
    my($fh, $n_types, $i, $blk) = @_;          # Return blocks in the block array ($blk) starting at index $i.
    my($time1, $blk1);

    my $lines_per_sat = blk_lines_per_sat($n_types);

    $blk1 = read_obs_block($fh, $lines_per_sat);
    while ( defined $blk1 && $blk1 ne "" ) {
        $blk->[$i] = $blk1;
        $i++;
        $blk1 = read_obs_block($fh, $lines_per_sat);
    }
    return $i;
}


sub read_obs_block {                # Read sat obs block of a RINEX OBS file from a file handle ($fh).
    my($fh, $lines_per_sat) = @_;   # Return obs block.
    my($year, $mon, $day, $hour, $min, $sec, $flag, $n_sats);
    my($blk, $time, $n);

    my $i = 0;
    my $n_lines;
    $_ = <$fh>;
    while ( defined $_ && $i < 50 ) {
	if ( m/ +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{0,2}\.[0-9]{1,}) +[0-9] +([0-9]{1,2})/ ) {
            ($year, $mon, $day, $hour, $min, $sec, $flag, $n_sats) = ($1, $2, $3, $4, $5, $6, $7, $8);
#	    print STDERR join " ", $year, $mon, $day, $hour, $min, $sec, $flag, $n_sats, "\n";
	    if ( $flag == 4 ) {
		# Skip header (comment) lines that are in the middle of the file
		print STDERR "Rinex::read_obs_block: Info, discarding intermediate header lines, FLAG=4.\n";
		while ( ($_ = <$fh>) =~ m/COMMENT/i ) { }
		next;
	    }
            $blk = $_;
            $i = 0;
            $n = 0;
            $n_lines = $n_sats * $lines_per_sat;
            while ( $n < $n_lines ) {
                $blk .= <$fh>;
                $n++;
            }
            $blk =~ s/\r\n/\n/g;            # if have DOS CR/LF line endings, delete CR
	    if ( $n == $n_lines ) {
                return $blk;
	    } else {
                carp "Rinex::read_obs_block: Error, incomplete sat block, discarded => $blk";
		return "";
	    }
        } else {
	    carp "Rinex::read_obs_block: Warning, expected time tag line, got $_";
            if ( $_ =~ m/RINEX FILE SPLICE/i ) {
		print STDERR "Rinex::read_obs_block: Info, discarding intermediate header lines, RINEX FILE SPLICE.\n";
	    }
	    $i++;
	}
        $_ = <$fh>;
    }

    if ( $i >= 50 ) {
        croak "Rinex::read_obs_block: Fatal error, cannot find valid time epoch line to start sat block";
    }
    return "";
}


sub fetch_sat_obs {    # Fetch all of the sat observations for a specified sat # into arrays, return # of obs.
    my($hdr, $blk, $sat_no, $time, $obs, $lli, $snr, $augment) = @_;
    my($time1, @obs1, @lli1, @snr1);

    my $n_types = $hdr->{'n_obs_types'};
    my $n;
    my $type_list = $hdr->{'obs_types'};

    my $i_obs = 0;
    for my $b (@$blk) {
	$n = fetch_blk_sat_obs($b, $n_types, $type_list, $sat_no, \$time1, \@obs1, \@lli1, \@snr1);
	if ( $n > 0 ) {
            if ( $n == $n_types ) {
		augment_sat_obs($type_list, $n, \@obs1) if $augment;
	        $time->[$i_obs] = $time1;         # time tag array (sec. past J2000)
                $obs->[$i_obs]  = [ @obs1 ];      # obs array containing values of L1, L2, P1, P2, C1, etc.
                $lli->[$i_obs]  = [ @lli1 ];      # loss of lock indicator array
                $snr->[$i_obs]  = [ @snr1 ];      # SNR array (for L1 & L2 only)
                $i_obs++;
            } else {
	        carp "Rinex::fetch_sat_obs: Bad sat obs line, $n not equal to $n_types.";
            }
        }
    }
    return $i_obs;
}


sub fetch_blk_time {
    my($blk) = @_;

    if ( $blk =~ m/ +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{0,2}\.[0-9]{1,})/ ) {
        my ($year, $mon, $day, $hour, $min, $sec_frac) = ($1, $2, $3, $4, $5, $6);
	my $sec = int($sec_frac);
	my $frac = $sec_frac - $sec;
        return DateTime::calsec($year, $mon, $day, $hour, $min, $sec, $frac);
    } else {
        croak "Rinex::fetch_blk_time: Error, bad time tag line in blk => $blk";	
    }
}


sub fetch_blk_sat_obs {    # Fetch the obs for a specific sat # from a block
    my($blk, $n_types, $type_list, $sat_no, $time, $obs, $lli, $snr) = @_;
    my($i, $j);

    $$time = fetch_blk_time($blk);
    my ($n_sats, $sat_list) = fetch_blk_sat_list($blk);
    my $i_sat = blk_index_of_sat_or_type($sat_no, $sat_list);
    if ( $i_sat < 0 ) {
	return $i_sat;
    } else {
        my $lines_per_sat = blk_lines_per_sat($n_types);
        my @line = split '\n', $blk;

	my $sat_str = "";
        my($j, $n_spaces);
	my $offset = $i_sat * $lines_per_sat + 1;
        for ($i = 0; $i < $lines_per_sat; $i++) {
	    $j = $i + $offset;
	    if ( defined $line[$j] ) {
                $n_spaces = 80 - length($line[$j]) - 1;     # a total kludge, RINEX v2 sucks.
		$line[$j] .= " " x $n_spaces;
	    } else {
		$line[$j] = " " x 80;
	    }
	    $sat_str .= $line[$j];
	}

        return unpack_sat_obs($sat_str, $type_list, $obs, $lli, $snr);
    }
}


sub blk_lines_per_sat {
    my($n_types) = @_;
    return int( ($n_types - 1) / 5) + 1;
}


sub unpack_sat_obs {
    my($line, $type_list, $obs, $lli, $snr) = @_;

    my @types = split ' ', $type_list;
    my($val, $kind, $freq);
    my $l_val = 16;
    my $i_obs = 0;

    for my $type (@types) {
        $kind = substr($type, 0, 1);
        $freq = substr($type, 1, 1);
	$freq--;
        $val = substr($line, $i_obs*$l_val, $l_val);

	if ( $kind eq 'L' ) {
            unpack_obs_L($val, $i_obs, $obs, $freq, $lli, $snr);
	} elsif ( $kind eq 'P' || $kind eq 'C' ) {
            unpack_obs_P_C($val, $i_obs, $obs, $freq, $lli);
	} elsif ( $kind eq 'S' || $kind eq 'D' ) {
            unpack_obs_S_D($val, $i_obs, $obs);
	} else {
            croak "Rinex::unpack_sat_obs: Do not recognize obs type, must be L#, P#, S# or D#";

	}
	$i_obs++;
    }
    return $i_obs;
}

sub unpack_obs_L {
    my($str, $i, $obs, $f, $lli, $snr) = @_;

    if ( $str =~ m/^\s*([-0-9 ]+\.[0-9]{3})([ 0-9])([ 0-9])\s*$/ ) {
        $$obs[$i] = $1;
	if ( $2 ne " " ) {
            $$lli[$f] = $2;
	} else {
            $$lli[$f] = NIL;
	}
	if ( $3 ne " " ) {
            $$snr[$f] = $3;
	} else {
            $$snr[$f] = NIL;
	}
    } elsif ( $str =~ m/^\s*$/ ) {
        $$obs[$i] = NIL;
        $$lli[$f] = NIL;
        $$snr[$f] = NIL;
    } else {
	croak "Rinex::unpack_obs_L: L1/L2 obs format doesn't match: $str.\n  Should be three decimals after decimal pt and two flags.";
    }
}

   
sub unpack_obs_P_C {
    my($str, $i, $obs, $f, $lli) = @_;

    if ( $str =~ m/^\s*([-0-9]+\.[0-9]{3})([0-9]?)([0-9]?)\s*$/ ) {
        $$obs[$i] = $1;
    } elsif ( $str =~ m/^\s*$/ ) {
        $$obs[$i] = NIL;
    } else {
	croak "Rinex::unpack_obs_L: P1/P2 obs format has changed.\n  Should be three decimals after decimal pt and an optional flag.";
    }
}

   
sub unpack_obs_S_D {
    my($str, $i, $obs) = @_;

    if ( $str =~ m/^\s*$/ ) {
        $$obs[$i] = NIL;
    } else {
        $$obs[$i] = $str;
    }
}

   
sub blk_index_of_sat_or_type {    # Find index of a particular sat # in the normalized sat list.
    my($sat_no, $sat_list) = @_;  # Or find index of a type (L1) in the ordered type list.
    my $i = index($sat_list, $sat_no);
    if ( $i >= 0 ) {
	$i = int($i/3);
    }
    return $i;
}


sub fetch_sat_obs_counts {    # Accumulate obs counts for all sats and return in hash where keys are sat #'s.
    my($blk, $sats) = @_;
    my($n_blk_sats, $blk_sat_list, @blk_sats, $sat_no);

    my $n_tot = 0;
    for my $b (@$blk) {
	($n_blk_sats, $blk_sat_list) = fetch_blk_sat_list($b);
	@blk_sats = split ' ', $blk_sat_list;
	for $sat_no (@blk_sats) {
	    $sats->{$sat_no}++;
	    $n_tot++;
	}
    }
    return $n_tot;
}


sub fetch_blk_sat_list {    # Fetch # of sats and normalized sat list from 1st line of obs block.
    my($blk) = @_;
    my($sat_list);

    if ( $blk =~ m/ +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{0,2}\.[0-9]{1,}) +[0-9] +([0-9]{1,2}).(.+)\s*\n/ ) {
        my ($year, $mon, $day, $hour, $min, $sec, $n_sats, $sats) = ($1, $2, $3, $4, $5, $6, $7, $8);
	$sat_list = normalize_sat_list($n_sats, $sats);
	return ($n_sats, $sat_list);
    } else {
	carp "fetch_blk_sat_list: Bad sat obs block => $blk";
	return ( 0 );
    }
}


sub normalize_sat_list {    # Converts list of sat numbers to format "03,09,24,32", leading zeros, comma separated
    my($n_sats, $rinex_str) = @_;
    my($sat_list, $i, $n);
    
    $sat_list = "";
    for ( $i=0; $i < $n_sats; $i++ ) {
	$sat_list .= sprintf("%2.2d ", substr($rinex_str, $i*3, 2));
    }
    return $sat_list;
}


sub merge_rinex_obs {    # Merge RINEX OBS from multiple files doing a proper time sort, and filter range of times.
    my($filter_begin_time, $filter_end_time, $out_fh, $files) = @_;
    
    my $first = 1;
    my $need_sort = 1;                                # need to do time sort if TRUE
    $need_sort = 0 if @$files == 1;                   # if only one input file, then already time sorted

    my $filter_time = 0;                              # filtering time range if TRUE
    my($filter_begin_sec, $filter_end_sec);           # begin & end of filtering time range
#    my($filter_begin_str, $filter_end_str);           # begin & end of filtering time range
    if ( $filter_begin_time ne "" || $filter_end_time ne "" ) {
	$filter_time = 1;
	$filter_begin_time = "1950/01/01,00" if $filter_begin_time eq "";
	$filter_end_time   = "2067/01/01,00" if $filter_end_time eq "";
        if ( $filter_end_time le $filter_begin_time ) {
	    croak "merge_rinex_obs: Error, filtering end time, $filter_end_time, before begin time, $filter_begin_time";
        }
        $filter_begin_sec = (new DateTime)->fromCalStr($filter_begin_time)->toSec;
        $filter_end_sec = (new DateTime)->fromCalStr($filter_end_time)->toSec;
#	$filter_begin_str = cal_time_str2obs_time_str($filter_begin_time);
#	$filter_end_str = cal_time_str2obs_time_str($filter_end_time);
    }

    my %hdr1 = ();         # header from first file
    my %hdr2 = ();         # headers from rest of files
    my @blk = ();          # block array
    my $i = 0;             # index into block and time arrays
    my $fn;                # current filename
    my $n_types;           # # of obs types, L1, L2, etc.

    # read obs blocks from all files
    while ( $fn = shift @{$files} ) {
        next if -z $fn;		# Prevent endless errors if file is zero length. MJR011227
	open(IN, $fn) || croak "Rinex::merge_rinex_obs: Cannot open file $fn";
	print STDERR "merge_rinex_obs: Reading ", $fn, " . . .\n";
	if ( $first ) {
	    $first = 0;
            read_obs_header(\*IN, \%hdr1);
	} else {
            read_obs_header(\*IN, \%hdr2);
	    if ( $hdr2{'site'} ne  $hdr1{'site'} ) {
		croak "Rinex::merge_rinex_obs: New file, $fn, is for different site than first file.";
	    }
	}
        $n_types = $hdr1{'n_obs_types'};
        $i = read_obs_blocks(\*IN, $n_types, $i, \@blk);
	close(IN);
    }
    my $n_obs_read = @blk;
    if ( $n_obs_read <= 0 ) {
	croak "merge_rinex_obs: No obs read.";
    }

    # sort obs by time, and filter out obs not in specified time range
    print STDERR "merge_rinex_obs: Sorting/filtering obs . . .\n" if $need_sort || $filter_time;
    my @blk2 = ();
    my %seen = ();
    if ( $filter_time ) {
	if ( $need_sort ) {
            # If you don't understand the code below, just read it backwards from the last line and
            # and consider what each line does.  (Don't you just love functional programming style?)
            # This kind of construct is actually quite fast in perl and uses less memory.
            @blk2 = map { ${$_->[1]} }                      # extract block from two-element array
                grep { ! $seen{$_->[0]}++ }                 # remove duplicate times (it's a perl idiom)
                sort { $a->[0] <=> $b->[0] }                # sort using time string as key
                grep { $_->[0] >= $filter_begin_sec && $_->[0] < $filter_end_sec }  # filter by time range
                map { [ fetch_blk_time($_), \$_ ] } @blk;    # form two-element array of time string and block
        } else {
            @blk2 = grep { fetch_blk_time($_) >= $filter_begin_sec &&
                           fetch_blk_time($_) <  $filter_end_sec } @blk;   # filter by time range
        }
        my $n_obs_after_filtering = @blk2;
        if ( $n_obs_after_filtering <= 0 ) {
	    croak "merge_rinex_obs: No obs satisfied time range.";
        }
    } else {
	if ( $need_sort ) {
            @blk2 = map { ${$_->[1]} }                      # extract block from two-element array
                grep { ! $seen{$_->[0]}++ }                 # remove duplicate times (it's a perl idiom)
                sort { $a->[0] <=> $b->[0] }                # sort using time string as key
                map { [ fetch_blk_time($_), \$_ ] } @blk;   # form two-element array of time string and block
        } else {
            @blk2 = @blk;                                   # no change since no sorting or filtering
        }
    }

    # write out updated header and obs blocks
    print STDERR "merge_rinex_obs: Writing merged obs . . .\n";
    my $hdr = update_obs_header($hdr1{'ALL'}, substr($blk2[0], 0, 26), substr($blk2[$#blk2], 0, 26));
    print $out_fh $hdr;
    for my $b (@blk2) {
        print $out_fh $b;
    }
    return 1;
}


sub cal_time_str2obs_time_str {
    my($cal_str) = @_;
    my($str);

    my($year, $mon, $day, $hour, $min, $sec, $frac) = (new DateTime)->fromCalStr($cal_str)->toCal;
    $year -= int($year/100)*100;
    $str = sprintf " %2.2d %2d %2d %2d %2d %10.7f", $year, $mon, $day, $hour, $min, $sec+$frac;
    return $str;
}


sub obs_time_str2hdr_time_str {
    my($obs_str) = @_;
    my($str);

    if ( $obs_str
         !~ m/ +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{1,2}) +([0-9]{0,2}\.[0-9]{1,})/ ) {
	croak "merge_rinex_obs: Bad obs time string => $obs_str";
    }
    my($year, $mon, $day, $hour, $min, $sec) = ($1, $2, $3, $4, $5, $6);
    $year = DateTime::year_4digit($year);
    $str = sprintf "  %4.4d    %2d    %2d    %2d    %2d   %10.7f",
                       $year, $mon, $day, $hour, $min, $sec;
    return $str;
}


sub read_obs {
    my($in_fh, $hdr, $blk) = @_;

    read_obs_header($in_fh, $hdr);
    my $n_types = $hdr->{'n_obs_types'};
    my $i = 0;
    $i = read_obs_blocks($in_fh, $n_types, $i, $blk);
    close(IN);
    return $i;
}


sub time_filter_obs {
    my($filter_begin_time, $filter_end_time, $blk) = @_;

    if ( $filter_begin_time ne "" || $filter_end_time ne "" ) {
	$filter_begin_time = "1950/01/01,00" if $filter_begin_time eq "";
	$filter_end_time   = "2067/01/01,00" if $filter_end_time eq "";
        if ( $filter_end_time le $filter_begin_time ) {
	    croak "Rinex::time_filter_obs: Error, filtering end time, $filter_end_time, before begin time, $filter_begin_time";
        }
	print STDERR "Rinex::time_filter_obs: Filtering obs from $filter_begin_time to $filter_end_time\n";
	my $filter_begin_str = cal_time_str2obs_time_str($filter_begin_time);
	my $filter_end_str = cal_time_str2obs_time_str($filter_end_time);

        my @blk2 = grep { substr($_, 0, 26) ge $filter_begin_str &&
                       substr($_, 0, 26) lt $filter_end_str } @$blk;   # filter by time range
	return \@blk2;
    } else {
	return $blk;
    }
}


sub write_obs {
    my($out_fh, $hdr, $blk, $augment, $connect) = @_;
    my($i);

    my $type_list = $hdr->{'obs_types'};
    $type_list = augment_type_list($type_list) if $augment;
    my @types = split ' ', $type_list;
    my $type_cols = join "      ", @types;
    print $out_fh "#Sat Time      " . $type_cols . "    LLI1 LLI2 SNR1 SNR2 Hour        Date_Time \n";

    my %sat_counts = ();
    my $n_tot_obs = fetch_sat_obs_counts($blk, \%sat_counts);
    if ( DEBUG ) {
	print STDERR "\nSat obs counts:\n";
	for my $sat (sort keys %sat_counts) {
	    printf STDERR "%s %d\n", $sat, $sat_counts{$sat};
	}
	print STDERR "Total obs: ", $n_tot_obs, "\n\n";
    }

    my @sats = sort keys %sat_counts;
    my($n_sat_obs, $time1, $hour, $date_time);
    my(@time, @obs, @lli, @snr);

    print STDERR "Writing obs for sat # . . .\n";
    for my $sat (@sats) {
        print STDERR "PRN ", $sat, "\n";
	if ( DEBUG ) { print STDERR "\n"; }
        $n_sat_obs = fetch_sat_obs($hdr, $blk, $sat, \@time, \@obs, \@lli, \@snr, $augment);
	connect_L_obs($type_list, \@time, \@obs) if $connect;
	write_sat_obs($out_fh, $sat, $n_sat_obs, \@time, \@obs, \@lli, \@snr);
    }
    print STDERR "Done.\n";
}


sub connect_L_obs {
    my($type_list, $time, $obs) = @_;
    my($i, $type1, $n_val, @val, @val2);

    my $il = 0;
    my @i_jump = ();
    my @i_all_jump = ();

    my $y_jump = Y_JUMP_HUGE;
    my @type = split ' ', $type_list;
    for ($i = 0; $i < @type; $i++) {
	if ( $type[$i] =~ m/L/i ) {
	    $n_val = extract_obs($i, $obs, \@val);

	    $type1 = $type[$i];
	    if ( $type1 eq "L1" || $type1 eq "L2" ) {
		$y_jump = Y_JUMP_L1;
            } elsif ( $type1 eq "LI" ) {
		$y_jump = Y_JUMP_LI;
#            } elsif ( $type1 eq "LC" ) {
#		$y_jump = Y_JUMP_LC;
            } else {
		$y_jump = Y_JUMP_HUGE;		    
            }
            $i_jump[$il] = ();
            @i_jump = Connect::simple($n_val, \@val, $type1, $y_jump, \@val2);
            push @i_all_jump, [ @i_jump ];
	    $n_val = replace_obs($n_val, \@val2, $i, $obs);
	    $il++;
	}
    }
#    if ( DEBUG ) {
    if ( 1 ) {
	my @t_jump = ();
        for ($il = 0; $il < @i_all_jump; $il++) {
	    if ( $type[$il] eq "L1" || $type[$il] eq "L2" ) {
                for ($i = 0; $i < @{$i_all_jump[$il]}; $i++) {
                    push @t_jump, "$i_all_jump[$il][$i]  $type[$il]";
                }
            }
	}
	print STDERR "\nSummary of phase jumps:\n Pt   Types\n---- -----";
	my @s_jump = sort { (split(' ', $a))[0]  <=> (split(' ', $b))[0] } @t_jump;
	my($line);
        my $i_pt;
        my $prev_i_pt = 0;
	while ( $line = shift @s_jump ) {
	    ($i_pt, $type1) = split ' ', $line;
	    if ( $i_pt == $prev_i_pt )  {
		print STDERR " ", $type1;
	    } else {
		print STDERR "\n", $line;
	    }
	    $prev_i_pt = $i_pt;
        }
	print STDERR "\n\n";
    }
}


sub extract_obs {
    my($i, $obs, $val) = @_;

    my $n_val = @$obs;
    for (my $j = 0; $j < $n_val; $j++) {
	$$val[$j] = $$obs[$j][$i];
    }
    return $n_val;
}

sub replace_obs {
    my($n_val, $val, $i, $obs) = @_;

    for (my $j = 0; $j < $n_val; $j++) {
	$$obs[$j][$i]  = $$val[$j];
    }
    return $n_val;
}


sub write_sat_obs {
    my($out_fh, $sat, $n_sat_obs, $time, $obs, $lli, $snr) = @_;

    my($time1, $hour, $date_time);

    for (my $i = 0; $i < $n_sat_obs; $i++) {
        $time1 = new DateTime($$time[$i]);
        $hour = $time1->toHoursInDay;
        $date_time = $time1->toCalStr;
        print $out_fh join " ", $sat, $$time[$i], @{$$obs[$i]}, @{$$lli[$i]}, @{$$snr[$i]}, $hour, $date_time, "\n";
    }
}

sub augment_type_list {
    my($list) = @_;
    return $list . " LI PI";
}


sub augment_sat_obs {
    my($type_list, $i_obs, $obs) = @_;

    my ($li, $pi) = (NIL, NIL);
    my $l1 = get_obs("L1", $type_list, $obs);
    my $l2 = get_obs("L2", $type_list, $obs);
    my $p1 = get_obs("P1", $type_list, $obs);
    my $p2 = get_obs("P2", $type_list, $obs);
    if ( $p1 eq NIL ) {
        $p1 = get_obs("C1", $type_list, $obs);
    }

    if ( $l1 ne NIL && $l2 ne NIL ) {
        $li = calc_iono_L($l1, $l2);
    }
    $$obs[$i_obs++] = $li;

    if ( $p1 ne NIL && $p2 ne NIL ) {
        $pi = calc_iono_P($p1, $p2);
    }
    $$obs[$i_obs++] = $pi;
}


sub get_obs {
    my($type, $type_list, $obs) = @_;
    my($i_obs, $val);

    if ( ($i_obs = blk_index_of_sat_or_type($type, $type_list)) >=0 ) {
	return $$obs[$i_obs];
    } else {
	return NIL;
    }
}


sub calc_iono_L {
    my($l1, $l2) = @_;

    my $li = ($l1 / F1 - $l2 / F2) * SEC2TECU;
    return $li;
}

sub calc_iono_P {
    my($p1, $p2) = @_;

    my $pi = ($p2 - $p1) / C * SEC2TECU;
    return $pi;
}

1;


