# Statistics package 
#
# Stats.pm -- Simple statistics class:  computes mean, sigma, min, max, rms.
#
# Author: Brian Wilson
#    @(#) Stats.pm     1.4     05/29/02

package Stats;
use strict;

# Implemented by accumulating an array of five numbers:
#   no of points, sum of values, sum of squares, min, and max.
# Methods:
#   add    -- add a data point to the accumulating stats
#   calc   -- compute the five statistics:  n, mean, std dev, min, max, rms
#   print  -- compute and print the stats on one line
#   reset  -- zero the accumulators
#   addm   -- add an array of data points to the accumulators (add multiple)

# See Stats_test.pl for example usage.
# Briefly:
#   my $s = new Stats;
#   for my $v ( @vals ) {
#       $s->add($v);
#   }
#   $s->print("Lying with statistics");
#      or 
#   ($n, $mean, $sigma, $min, $max, $rms) = $s->calc();
#
#   Willy Bertiger 4/26/02 Modify algorithm for better numerical accuracy
#

sub new { bless( [0, 0.0, 0.0, 0.0, 0.0], shift ) }

sub add {
    my($self, $val) = @_;

    $self->[0]++;
    my $dval = $val - $self->[1];  # difference from current mean
    $self->[1] += $dval/$self->[0]; # update mean
    my $ddval = $val - $self->[1];
    $self->[2] += $dval*$ddval;

    if ( $self->[0] == 1 ) {
	$self->[3] = $val;
	$self->[4] = $val;
    } else {
        $self->[3] = $val if $val < $self->[3];
        $self->[4] = $val if $val > $self->[4];
    }
}

sub calc {
    my $self = shift;
    my($n, $mean, $sumsq, $min, $max) = @{$self};
    my($sigma, $rms) = (0.0, 0.0);

    if($n == 0){
        return (0,0,0,0,0,0);
    }
    if ($n < 2) { $sigma = 0 }
    else{
        my $sd2 = $sumsq/($n-1);
        $sigma = ($sd2 <= 0) ? 0 : sqrt($sd2);
    }
    $rms = sqrt($mean*$mean + $sumsq/$n);

    ($n, $mean, $sigma, $min, $max, $rms);
}

sub print {
    my ($self, $label) = @_;
    my($n, $mean, $sigma, $min, $max, $rms) = $self->calc();

    $label = "Stats" unless defined $label;
    print "$label:  ";
    print join "  ", $n, $mean, $sigma, $min, $max, $rms, "\n";
}

sub reset { 
    my $self = shift;
    
###    @{$self} = [0, 0.0, 0.0, 0.0, 0.0];
    for my $i (0 .. 4) {
	$self->[$i] = 0;
    }
}

sub addm {
    use Carp;
    my ($self, $vals_ref) = @_;

    carp "Stats->addm method:  Single arg must be an array reference" unless ref $vals_ref;
    for my $val ( @{$vals_ref} ) {
	$self->add($val);
    }
}

1;





