#!/usr/local/bin/perl -w 
# file join.pl

use strict;
use POSIX qw(tmpnam);
use Cwd;
use constant USAGE_MSG => <<USAGE;

Usage: join.pl -f file1_dump.txt -k 1 3 -c 4 'c6/100.' -f file2_dump.txt -c 4 8  [more options . . .]

Purpose: Join two files (after optional sorting) on *multiple* key columns, and output only the key
         columns and the specified (-c) data columns.

Options:
    -f, -file  <filename>      Name of text file to join and compare.  Must appear exactly twice to specify
                               the two (or more) files to compare.
    -k, -keys  <column #'s>    Column numbers to use as keys in the "joining" the two files (values in output).
                               If -k appears only once, then key columns for second file are assumed to be
                               the same as for the first file.  The keys should be specified in the order
                               of the existing or desired sort (for the join).
                               The keys can be specified as keydefs that will be passed to the Unix sort
                               function; e.g., 3n for numeric sort or 4d for dictionary sort.

    -c, -columns <column #'s>  Column numbers of fields (from the nth file) that you want to append to each line
                               in the output file.  The columns apply to the file specified on a preceding -f
                               switch.  Each specifer can also be an expression like the cl script.

    -s, -sort                  Sort the files before joining (i.e., assume they are not sorted on input).
    -t, -tolerance             Tolerance value for comparing numeric fields and still considering them equal.
    -m, -match <regexp>        Only process lines (from the input files) that match the regular expression.
    -v, -verbose               Echo activities if verbose mode.
    -z                         Null switch to end list of args passed to -k and -c switches.

Explanation of example usage:
    -f file1_dump.txt => Use this file as the first file in the comparison.
    -k 1n 3           => Use columns 1 and 3 from each file as keys for the join.  The values of these
                         columns will appear as the first two columns of the output text file.
                         If -sort is specified, then the sort will be numeric order on column 1 and 
                         lexicographic order on column 3.
    -c 4 'c6/100.'    => Select column 4 (of the first file) and an expression computed from column 6
                         as two data columns to append to each line of the output file.
    -f file2_dump.txt => Use this file as the second file in the comparison.
    -c 4 8            => Select columns 4 and 8 (of the second file) to append to each line of the
                         output file.

    The output file will contain six columns:  the two keys, the two specified columns from file1,
    and the two specified columns from file2.  Presumably, columns 3 and 5 of the output file contain
    comparable fields from the two files, and can now be plotted or differenced for comparison.

USAGE

my ($keys1, $keys2, $file1, $file2, $cols1, $cols2, $regexp, $tolerance);
my $sort_flag = 0;
my $numeric_flag = 0;
my $verbose = 0;

while ( $_ = shift @ARGV ) {
    if ( m/^-k/ ) {
	if ( ! defined $keys1 ) {
	    $keys1 = get_arg_vals();
	} elsif ( ! defined $keys2 ) {
	    $keys2 = get_arg_vals();
	} else {
	    die USAGE_MSG;
	}
    } elsif ( m/^-f/ ) {
	if ( ! defined $file1 ) {
	    $file1 = shift @ARGV;
	    die "join.pl: File does not exit: $file1" unless -e $file1;
	} elsif ( ! defined $file2 ) {
	    $file2 = shift @ARGV;
	    die "join.pl: File does not exit: $file2" unless -e $file2;
	} else {
	    die USAGE_MSG;
	}
    } elsif ( m/^-c/ ) {
	if ( ! defined $cols1 ) {
	    $cols1 = get_arg_vals();
	} elsif ( ! defined $cols2 ) {
	    $cols2 = get_arg_vals();
	} else {
	    die USAGE_MSG;
	}
    } elsif ( m/^-s/ ) {
	$sort_flag = 1;
    } elsif ( m/^-t/ ) {
	$tolerance = shift @ARGV;
    } elsif ( m/^-v/ ) {
	$verbose = 1;
    } elsif ( m/^-m/ ) {
	$regexp = shift @ARGV;
    } elsif ( m/^-z/ ) {
        # no action, just eat null switch
    } else {
	die "join.pl: Illegal switch => $_";
    }
}

die USAGE_MSG unless (defined $file1 && defined $file2 && defined $cols1 && defined $cols2 && defined $keys1);

$keys2 = $keys1 if ! defined $keys2;
my $n_keys1 = scalar( split(' ', $keys1) );
my $n_keys2 = scalar( split(' ', $keys2) );
die "join.pl: Must have same number of key columns for both files." if ( $n_keys1 != $n_keys2 );

my $key_cols1 = col_numbers($keys1);  # extract column numbers only from keydef specifiers
my $key_cols2 = col_numbers($keys2);
$cols1 = quote_expr($cols1);          # requote expr for cl call
$cols2 = quote_expr($cols2);

my $tmpdir = tmpnam();
mkdir($tmpdir, 0770) or die "join.pl: unable to make temp directory: $tmpdir";
my $file1o = "$tmpdir/file1o";
my $file2o = "$tmpdir/file2o";

# Accumulate key and data columns into tmp files to prepare for sort and join.
if ( defined $regexp ) {
    print STDERR "grep \'$regexp\' $file1 | cl $key_cols1 $cols1 > $file1o\n" if $verbose;
    print STDERR `grep \'$regexp\' $file1 | cl $key_cols1 $cols1 > $file1o`;
    print STDERR "grep \'$regexp\' $file2 | cl $key_cols2 $cols2 > $file2o\n" if $verbose;
    print STDERR `grep \'$regexp\' $file2 | cl $key_cols2 $cols2 > $file2o`;
} else {
    print STDERR "cat $file1 | cl $key_cols1 $cols1 > $file1o\n" if $verbose;
    print STDERR `cat $file1 | cl $key_cols1 $cols1 > $file1o`;
    print STDERR "cat $file2 | cl $key_cols2 $cols2 > $file2o\n" if $verbose;
    print STDERR `cat $file2 | cl $key_cols2 $cols2 > $file2o`;
}

my $file1s = "$tmpdir/file1s";
my $file2s = "$tmpdir/file2s";
my @keys = split ' ', $keys1;
my $order;

# Sort both files if asked to.
my $sort_keys = "";
my $cmp_keys = "";
for my $i (1 .. $n_keys1) {
    ($order) = $keys[$i-1] =~ m/^[0-9]+([^0-9]+)/;
    $order = "" if ! defined $order;
    $cmp_keys .= " -k ${i}$order";

    $order = "n" if $order eq "t";
    $sort_keys .= " -k ${i}$order";
}
if ( $sort_flag ) {
    print STDERR "sort $sort_keys $file1o > $file1s\n" if $verbose;
    print STDERR `sort $sort_keys $file1o > $file1s`;

    print STDERR "sort $sort_keys $file2o > $file2s\n" if $verbose;
    print STDERR `sort $sort_keys $file2o > $file2s`;
} else {
    rename( $file1o, $file1s );
    rename( $file2o, $file2s );
}
$cmp_keys .= " -t $tolerance" if defined $tolerance;

# Join the files.
print STDERR "join_simple.pl $cmp_keys $file1s $file2s\n" if $verbose;
             `~/perllib/join_simple.pl $cmp_keys $file1s $file2s`;

# Cleanup tmp files.
my $cwd = cwd();
chdir $tmpdir;
#unlink $file1o, $file2o, $file1s, $file2s;
chdir $cwd;
unlink $tmpdir;

exit 0;


# Subroutines follow.

sub get_arg_vals {
    my $vals = "";

    while ( defined ($_ = shift @ARGV) && ! m/^-/ ) {
	$vals .= "$_ ";
    }
    unshift @ARGV, $_ if defined $_;
    return $vals;
}

sub col_numbers {
    my $str = shift @_;
    my $str2 = "";

    foreach my $part (split ' ', $str) {
        if ( $part =~ m/^([0-9]+)/i ) {
	    $str2  .= "$1 ";
	} else {
            die "join.pl: Bad key column specifier: $str";
	}
    }
    return $str2;
}

sub quote_expr {
    my $str = shift @_;
    my $str2 = "";

    foreach my $part (split ' ', $str) {
        if ( $part =~ m/^[0-9]+$/ ) {
	    $str2  .= "$part ";
	} else {
            $str2  .= "'$part' ";
	}
    }
    return $str2;
}

