#!/usr/local/bin/perl -w 
# file join_simple.pl

use strict;
use Cwd;
use constant USAGE_MSG => <<USAGE;

Usage: join_simple.pl -k 1n -k 2d -k 3 file1 file2

Purpose:  Join two (sorted) files on *multiple* key columns, and output the key columns once and
          the data columns from both files.

Options:
    -k, -key <keydef or #>   Column numbers to use as keys in the "joining" the two files.
                             The keys should be specified in the sorted order, and they can
                             be specified as keydefs like those passed to the Unix sort
                             function; e.g., 3n for numeric sort or 4d for dictionary sort.

    -t, -tolerance           Tolerance value for comparing numeric fields and still considering them equal.

Notes:  See also join.pl for more functionality.

USAGE

my $keys_str = "";
my $tol = 0.0;
while ( ($_ = shift @ARGV) && m/^-/ ) {
    if ( m/^-k/ ) {
        $keys_str .= shift @ARGV;
        $keys_str .= " ";
    } elsif ( m/^-t/ ) {
	$tol = shift @ARGV;
    }
}
my $file1 = $_;
my $file2 = shift @ARGV;
die USAGE_MSG unless (defined $file1 && defined $file2 && $keys_str ne "");

open(F1, $file1) or die "join_simple.pl: Cannot open file1: $file1";
open(F2, $file2) or die "join_simple.pl: Cannot open file1: $file2";

# Construct comparison expression that will be eval'ed for each line of both files.
# Returns -1 for lt, 0 for equal, and +1 for gt.
# Example expression for args of:  -k 1n -k 2d -k 3
#   $cols1[0] <=> $cols2[0] ||
#   $cols1[1] cmp $cols2[1] ||
#   $cols1[2] cmp $cols2[2]
#
my @keys = split ' ', $keys_str;
my @key_cols = ();
my $n_keys = scalar(@keys);
my $cmp_expr = "";
my ($col, $order);
for my $key (@keys) {
    ($col, $order) = $key =~ m/^([0-9]+)([^0-9]*)/;
    $col--;
    push @key_cols, $col;
    $cmp_expr .= " || \n" if $cmp_expr ne "";
    if ( $order =~ m/(n|t)/i ) {
        if ( $order =~ m/n/i ) {
            $cmp_expr .= "\$cols1[$col]";
            $cmp_expr .= " <=> ";
            $cmp_expr .= "\$cols2[$col]";
        } else {
            $cmp_expr .= "cmp_num(\$cols1[$col], \$cols2[$col], $tol)"
	    }
    } else {
        $cmp_expr .= "\$cols1[$col]";
	$cmp_expr .= " cmp ";
        $cmp_expr .= "\$cols2[$col]";
    }

}

my $line1 = <F1>;
die "join_simple.pl: Empty file1: $file1" if ! defined $line1;
my $line2 = <F2>;
die "join_simple.pl: Empty file2: $file2" if ! defined $line2;
my $n_keysm1 = $n_keys - 1;

# Prepare non-key column list for both files (use in slices below).
my @cols1 = split ' ', $line1;
my @cols2 = split ' ', $line2;
my $n_cols1 = scalar(@cols1);
my $n_cols2 = scalar(@cols2);
my @cols = ints_0_up_to_n($n_cols2);
my @non_key_cols2 = complement_of(\@key_cols, \@cols);

my ($line, $cmp, $i, $j);
while ( defined $line1 && defined $line2) {
    $cmp = eval $cmp_expr;
    if ( $cmp <= 0 ) {
	if ( $cmp == 0 ) {
            &construct_out_line();
            print STDERR $line, "\n";
	}
	$line1 = <F1>;
        @cols1 = split ' ', $line1 if defined $line1;
    } else {
	$line2 = <F2>;
        @cols2 = split ' ', $line2 if defined $line2;
    }
}

exit 0;


# Subroutines follow.

sub cmp_num {
    my($n1, $n2, $tol) = @_;

    if ( abs($n2 - $n1) < $tol ) {
	return 0;
    } elsif ( $n1 < $n2 ) {
	return -1;
    } else {
	return 1;
    }
}
	
sub construct_out_line {
    $line = join " ", @cols1;
    $line .= " ";
    $line .= join " ", @cols2[@non_key_cols2];
}    

sub complement_of {
    my($rsubset, $rset) = @_;
    my @cmpl = ();

    for my $elt (@$rset) {
	if ( ! element_of($elt, $rsubset) ) {
	    push @cmpl, $elt;
	}
    }
    return @cmpl;
}

sub element_of {
    my($i, $relts) = @_;
    
    for my $elt (@$relts) {
	if ( $i == $elt ) {
	    return 1;
	}
    }
    return 0;
}

sub ints_0_up_to_n {
    my $n = shift @_;
    my $i = 0;
    my @ints;
    
    while ( $i < $n ) {
	push @ints, $i;
        $i++;
    }
    return @ints;
}
