#!/usr/local/bin/perl -w 
#
# rdsmsol2cmpr_coefs -- Read GIM coefficient parameter values from smsol*.txt file and
#                       reformat them into new compressed *.gim format
#

use strict;
use lib "$ENV{'HOME'}/perllib";
use DateTime;

if ( @ARGV != 1 ) {
    print << "EOS";
Usage: smsol2cmpr_coefs  <rdsmsol.txt file>

   *** Writes GIM coefficients in gim*.dat format to STDOUT. ***

EOS
exit 1;
}

my $fn = shift @ARGV;
open(IN, $fn) || die "rdsmsol2cmpr_coefs: Fatal error, could not open rdsmsol file $fn";

my ($line, $line1, $line2, $line3);
my ($coef, $ch_time, $epoch, $ch_epoch);
my (@param, $ip, $np, $it, $nt);
my (@delta, $delta1, @val, $val1, @sigma, $sigma1);

$ip = -1;                                   # counter for number of coef params
while ( $line = <IN> ) {
  # find all parameter blocks that are for coef parameters
  if ( $line =~ m/STOCHASTIC PARAMETER:\s+(\S+)\^\s*EPOCH:\s+(\S+ \S+)/ ) {
      $coef = $1;  $ch_time = $2;
      next if $coef =~ m/BIAS/;

      if ( $coef !~ m/(SPC2|TRIN|TAVG)P *([0-9]+)/ ) {
	  die "rdsmsol2cmpr_coefs: Error, unknown coef param type => $1";
      }

      $line1 = $line;                       # save line for error msgs
      $ip++;
      if ( $ip == 0 ) {                     # if first parameter
	  $ch_epoch = $ch_time;             # then save epoch time
	  $epoch = (new DateTime)->fromCh($ch_time);
      } else {
	  if ( $ch_time ne $ch_epoch ) {    # else match epoch to first one, must be equal
	      die "rdsmsol2cmpr_coefs: Error, epoch time does not match other coef parameters =>\n $line";
          }    
      }

      $line2 = <IN>;
      if ( $line2 !~ m/OASIS NAME: (.*)\s*NOMINAL/ ) {
	  die "rdsmsol2cmpr_coefs: Bad or missing line in rdsmsol file after =>\n$line1\n$line2";
      }
      $param[$ip] = $1;
      <IN>;  <IN>;          # skip two lines

      # read all times in parameter block
      $it = -1;                             # counter for number of times
      while ( ($line3 = <IN>) !~ m/^\s*$/ ) {      # read until blank line
	  ($delta1, $val1, $sigma1) = split ' ', $line3;
	  $it++;
	  if ( $ip == 0 ) {                    # if first parameter
              $delta[$it] = $delta1;           # then save stochastic update times as delta from epoch
          } else {
              if ( $delta1 != $delta[$it] ) {  # else make sure deltas are same for this parameter
                  die "rdsmsolcoefs: Error, time mismatch for parameter =>\n$line1\n$line2\n$line3";
              }
          }
	  $val[$ip][$it] = $val1;
	  $sigma[$ip][$it] = $sigma1;
      }
      if ( $ip == 0 ) {                     # if first parameter
	  $nt = $it + 1;                    # then save number of stochastic updates
      } else {
	  if ( $nt != $it + 1 ) {           # else make sure number is the same for this parameter
              die "rdsmsol2cmpr_coefs: Error, mismatched number of times for parameter =>\n$line1\n$line2";
          }
      }
  }    
}
$np = $ip + 1;                              # save number of coef parameters

# round times down to minute by subtracting # of seconds in first epoch
my ($year, $mon, $day, $hour, $min, $sec, $frac) = $epoch->toCal;
my $sec_shift = $sec;
print "! This is rdsmsol2cmpr_coefs output. Multiple epoch times as noted.\n";
if ( $sec_shift ) { printf "! The epoch has been SHIFTED by %5d seconds\n", $sec_shift }

my $beg_time = new DateTime;
my $end_time = new DateTime;
my $epoch_sec = $epoch->toSec - $sec_shift;     # base epoch has been shifted

my $last_param = fix_param_name( $param[$np-1] );
my $param_type = substr($last_param, 0, 4);
my $n_layers = substr($last_param, 4, 1);
$n_layers = 1 if $n_layers !~ m/[0-9]/;
my $n_params = substr($last_param, 6, 3);
if ( $np != $n_params * $n_layers ) {
    die "rdsmsol2cmpr_coefs: Error, mismatched number of params & layers =>\nnp = $np\n$last_param\n";
}

# loop over times and print start and end time for each stochastic update
for ( $it = 0; $it < $nt; $it++ ) {
    if ( $it == 0 ) {
        $beg_time->fromSec( $epoch_sec + $delta[$it] );      # first time so don't add 1 sec.
    } else {
        $beg_time->fromSec( $epoch_sec + $delta[$it] + 1 );
    }
    if ( $it < $nt - 1 ) {
	$end_time->fromSec( $epoch_sec + $delta[$it+1] );
    } else {                                                 
	$end_time->fromCalStr("2067/12/31,00:00:00");        # last time is forever   
#	$end_time->fromSec( $epoch_sec + 2*$delta[$it] - $delta[$it-1] );
    }
    printf "%s %s %s %s %4d %2d\n", "TIME", short_date_time_str($beg_time),
           short_date_time_str($end_time), $param_type, $n_params, $n_layers,  "\n";
#    printf STDERR "%s %s %s %s %4d %2d\n", "TIME", short_date_time_str($beg_time),
#           short_date_time_str($end_time), $param_type, $n_params, $n_layers,  "\n";

    # loop over param names and print param name, estimate and sigma
    my $i_layer = 0;
    my $tmp;
    for ( $ip = 0; $ip < $np; $ip++ ) {
        printf "%11.5f %7.2f", $val[$ip][$it], $sigma[$ip][$it];
	$tmp = ($ip+1) / $n_params;
        if ( $tmp == int($tmp) ) {
            $i_layer++;
            printf " %2d", $i_layer;
        }
        printf "\n";
    }
}
print "STOP\n";
exit 0;

# subroutines below
sub short_date_time_str {
    my($date_time) = @_;
    my ($year, $mon, $day, $hour, $min, $sec, $frac) = $date_time->toCal;

    sprintf "%4.4d%2.2d%2.2d-%2.2d%2.2d%2.2d", $year, $mon, $day, $hour, $min, $sec;
}

sub fix_param_name {
    my($str) = @_;
    my($fix, $coef_type, $n);

    if ( $str =~ m/(....)P *([0-9]+)/ ) {   # coef type must be 4 chars, e.g. SPC2, TRIN, TAVG
	($coef_type, $n) = ($1, $2);
        $fix = $coef_type;                  # start with coef type, e.g. SPC2, TRIN, or TAVG
        if ( length($n) == 3 ) {            # original signal shell GIM, SPC2P243 -> SPC2 0243
            $fix .= " 0" . $n;
        } elsif ( length($n) == 4 ) {           # multishell GIM, SPC2P1243 -> SPC21 243
            $fix .= substr($n, 0, 1) . " " . substr($n, 1, 3);
        } else {
            die "rdsmsol2coefs: Error, cannot understand coef param numbers => $str";
        }
    } else {
        die "rdsmsol2coefs: Error, cannot understand coef param names => $str";
    }
    return $fix;
}
