#!/usr/local/bin/perl -w

use strict;
use lukeutils;
use statistics;
use stlocs;

$|=1;

################################################################################
package Tec_Dump;
################################################################################
#Provides convenient functions to manipulate, sort, and analyze tec_dump files.
################################################################################

##Arguments: "full path" <optional>
sub new 
 {
  my @tec_dump_lines;  
  my $self = bless ( \@tec_dump_lines, shift );
  load ($self,@_);
  return $self;
 }

##Arguments: "full path"
sub load
 {
  use Carp;

  my ($self, $filename)=@_;
  my @w;
  my @error_lines;
  my ($used_data, $rejected_data)=(0,0);
 
  if ($filename)
   {
    open (FILER,$filename) || croak "Can't open $filename!\n";
    foreach (<FILER>)
     { 
      if (substr($_,0,1) ne "#")   #Ignore pounded comments
       {
        @w=split(' ',$_);
        if ($w[18]<0) { $rejected_data++; } else {$used_data++; } #count data types
        if ($w[13]<1) #impossibly low STEC value
         { push @error_lines,$_; }
#        else   #Used to be excluded in run, but now we include all data for analysis
         { push @$self,$_; }
       }
     }
#    @$self=@{lukeutils::file_to_array($filename,$warning[0])};
    foreach (@$self) {@w=split(' ',$_); $_=$_."  ".($w[17]/$w[13]*100);} #Add percent STECDiff to end
   }
  return ($used_data+$rejected_data,$used_data,$rejected_data,\@error_lines);
 }

##Arguments: none
sub count
 {
  my $self=shift;
  return scalar(@$self);
 }

##Arguments: none
sub sprint
 {
  my $self = shift;
  my $s;
  foreach (@$self) { $s.=sprintf("$_\n"); }
  return $s;
 }

##Force the order of the tec_dump array to match a given station list ordering, else default
sub Force_Order
 {
  my ($self, $rOrder)=@_;
  if (scalar(@$rOrder)>0) {lukeutils::sort_like_this($self,$rOrder);}
  else                    {Tec_Dump::sort($self,2);}
 }

#Sorts a tec_dump-style list based on any passed keys
#Arguments: string array of columns to sort by
#NOTE: Columns are specified as 1-based, not 0-based
#####
sub sort
 {
  my $self = shift;

  #Build appropriate Schwartzian transform
  my $worker;
  my @comparison_list;
  #Prepare an array of appropriate comparison operators for the columns requested (string vs. numeric)
  for (my $t=0; $t<=$#_; $t++) 
  {
   my @array=split(' ',$$self[0]);
   if (lukeutils::valid_number($array[$_[$t]-1]))
    {$comparison_list[$t]="<=>"}
   else
    {$comparison_list[$t]="cmp";}
  }

  $worker ="map { \$_->[0] } sort {";
  for (my $t=0; $t<=$#_; $t++) { $worker.=" \$a->[" . ($t+1) . "] ".$comparison_list[$t]." \$b->[" . ($t+1) . "]"; if ($t<$#_) {$worker.=" ||";} }
  $worker.="} map { [ \$_, (split ' ', \$_)[";
  for (my $t=0; $t<=$#_; $t++) { $worker.= $_[$t]-1; if ($t<$#_) {$worker.=",";}}
  $worker.="] ] } \@\$self;";

#  print $worker,"\n";

  @$self=eval($worker);

#  map  { $_->[0] }
#  sort { $a->[1] cmp $b->[1] ||
#         $a->[2] cmp $b->[2] ||
#         $a->[3] <=> $b->[3]    }
#  map  { [ $_, (split ' ', $_)[1,2,17] ] } @$self;
 }

#While maintaining the order of the first column argument,
#sorts within the entries of that column based on any number of passed columns
#Thus, for column 1 of a tec_dump file, can preserve current station ordering
#while sorting by stec_diff, satellite, etc.
#Arguments: string array of columns to sort by, started by column to preserve
#NOTE: Columns are specified as 1-based, not 0-based
#####
sub Sort_Within_Subsection
 {
  my $self     = shift;
  my $preserve = shift;
  my @w=split(" ",$$self[0]);
  my @temp;
  
  my ($left,$last)=(0,$w[$preserve-1]);
  for (my $t=0;$t<@$self;$t++)
   {
    @w=split(" ",$$self[$t]);
    if ($last ne $w[$preserve-1])
     {
      @temp=@$self[$left..($t-1)];
      Tec_Dump::sort(\@temp,@_);
      @$self[$left..($t-1)]=@temp;
      $left=$t;
      $last=$w[$preserve-1];
     }
   }
 }

##Produce statistics on any sites that contain negative TEC or negative STEC Error measurements
#sub error_lines_report
# {
#  my $self=shift;
#  my @worker;
#  my %stats_stec;
#  my %stats_error;
#  my $result;
#  my @order;#

#  $result="STAT 

#  foreach (@$self)
#   {
#    @worker=split(' ',$_);
#    if (!defined $stats_stec_diff{$worker[1]}) 
#     { 
#       $stats_stec{$worker[1]}=new Statistics;
#      $stats_error{$worker[1]}=new Statistics; 
#       $stats_stec{$worker[1]}->add($worker[13],$worker[2]);
#      $stats_error{$worker[1]}->add($worker[18],$worker[2]);
#      @order=(@order,$worker[1]);
#     }
#   }

#  foreach (@order) 
#   {
#    my ($n, $mean, $sigma, $min, $max, $rms, $minlabel, $maxlabel)=$stats{$_}->calc;
#    $result.="$_: $stats{$_} lines\n"; 
#   }

#  return $result;
# }

#Return an array of lines that exceed a certain magnitude +/- in the specified column
#Does NOT remove them from the original file. Column is 1 based.
####
sub return_outliers
 {
  my ($self, $col, $value)=@_;
  my @worker;
  my @answer;

  foreach (@$self)
   {
    @worker=split(' ',$_);
    if (abs($worker[$col-1])>=abs($value)) { push @answer, $_; }
   }
  return @answer;
 }

#Apply a filter to remove stations with an STECDIFF% within user-specified
#threshholds which are latitude dependent. "t" threshholds are percentage, while "a"
#threshholds are absolute. Latitude ranges are abs(0-30), abs(30-60) and abs(60-90).
#This routine also tacks on a new column on the end of the tec_dump lines which records
#if this line violates the "A"bsolute, "P"ercentage, or "B"oth threshholds.
####
sub filter_stec_diff_percent_threshhold_outliers
 {
  my ($self, $t1, $t2, $t3, $a1, $a2, $a3, $stlocs)=@_;
  my (@worker,@result);
  my @lookup = ("P","A","B");
 
  my $code;
  my $threshhold;
  my $abser;
  my $bare; #this is the minimum absolute STEC diff that must exist before percentages will be counted
  foreach (@$self)
   {
    @worker=split(' ',$_);
    my ($station,$lat,$percent_error,$abs_error)=($worker[1],$stlocs->latitude($worker[1]),abs($worker[34]),abs($worker[17]));
    if    ($lat<30) {$threshhold=$t1; $abser=$a1; $bare=10;}
    elsif ($lat<60) {$threshhold=$t2; $abser=$a2; $bare=7 ;}
    else            {$threshhold=$t3; $abser=$a3; $bare=5 ;}
    $code=0;

    if (($percent_error>=$threshhold)&&($abs_error>$bare)) {$code+=1;}
    if  ($abs_error    >=$abser     )                      {$code+=2;}
    if ($code) {$worker[35]=$lookup[$code-1]; @result=(@result,join(" ",@worker));}
   }
  @$self=@result;
 }

#Determine whether sites in the list are DSN sites looking towards
#the equator. If the list has already been processed to remove
#non-extrema, this will determine DSN run acceptability.
###
sub check_for_bad_DSN
 {
  my $self=shift;
  my ($madr_flag, $gold_flag, $tidb_flag, @temper, $az);

  $madr_flag=$gold_flag=$tidb_flag="";
  foreach (@$self)
   {
    @temper=split(' ',$_);
    if ($temper[18]>=0) {  #only if a used point by GIM
       if ((grep /MADR|MAD2/i, $temper[1])&&($temper[35] eq "B")) { $madr_flag="MADR "; }
    elsif ((grep /GOLD|GOL2/i, $temper[1])&&($temper[35] eq "B")) { $gold_flag="GOLD "; }
    elsif ((grep /TIDB|TID2/i, $temper[1])&&($temper[35] eq "B")) { $tidb_flag="TIDB "; }
    }
   }
  return $madr_flag.$gold_flag.$tidb_flag;
 }

#Filter a tec_dump such that only the most positive and negative entries
#for each station-satellite combination are reported.
#Arguments: number of points to include maximum for each satellite/station/violation-type trio, dflt 1
#####
sub filter_extrema
 {
  my ($self, $num) = @_;
  if (!defined $num) {$num=1;}

  Tec_Dump::sort($self,2,3,36,35);
  my ($last,$count);
  my (@worker,@filtered,$temp);

  $last=" ";
  for (my $t=0;$t<scalar(@$self);$t++)     #look for negative minima
   {
    @worker=split(' ',$$self[$t]);
    $temp=$worker[1].$worker[2].$worker[35]; #form the Station, Sattelite, and Threshhold Violation trio
    if ($worker[17]<0)
     {
      if ($temp ne $last) { $last=$temp; $count=1; } else {$count++;}
      if ($count<=$num) { @filtered=(@filtered, $$self[$t]); }
     }
   }
  $last=" ";
  for (my $t=scalar(@$self)-1;$t>=0;$t--)     #look for positive maxima
   {
    @worker=split(' ',$$self[$t]);
    $temp=$worker[1].$worker[2].$worker[35]; #form the Station, Sattelite, and Threshhold Violation trio
    if ($worker[17]>0)
     {
      if ($temp ne $last) { $last=$temp; $count=1; } else {$count++;}
      if ($count<=$num) { @filtered=(@filtered, $$self[$t]); }
     }
   }

  @$self=@filtered;
  Tec_Dump::sort($self,2,3,36,35);
 }

#This subroutine outputs useful columns in a pretty display.
#Requires a reference to a current Stlocs object.
#Arguments: stlocs_object
#Control is a string that can switch in and out certain behavior
#"threshhold" will cause sprint_nice to report column 35, which is the threshhold violated
#Column 35 is added by filter_stec_diff_percent_threshhold_outliers.
#####
sub sprint_nice
 {
  my ($self, $stlocs, $rCodes, $control, $label) = @_;
  my @w;
  my $output;
  my $threshhold_flag=0;

  if ($control) {if (grep "threshhold",$control) {$threshhold_flag=1;}}

  $output.=sprintf("%4s CU %3s %6s %5s %5s %4s %5s %5s %5s %6s %6s %5s %8s %5s\n",
    "site","sat","resid ","  % ","STEC","err"," az","  el ","IPLat","IPLon","MgLat","MgLon","ddhhmmss","ltime");
  if ($threshhold_flag) {$output=lukeutils::trim($output)." T\n";}
  $output.=sprintf("%4s -- %3s %6s %5s %5s %4s %5s %5s %5s %6s %6s %5s %8s %5s\n",
    "-"x4 ,"-"x3,"-"x6   ,"-"x5 ,"-"x5 ,"-"x4,"-"x5,"-"x5  ,"-"x5  ,"-"x6  ,"-"x6  ,"-"x5  ,"-"x8   ,"-"x5);
  if ($threshhold_flag) {$output=lukeutils::trim($output)." -\n";}
  my $valid_codes = scalar(keys(%$rCodes))>0;
  my $coder;
  foreach (@$self)
  {
   @w=split(' ',$_);
   if ($valid_codes) {$coder=$$rCodes{$w[1]};} else {$coder=" ";}
   if ($w[18]<0) { $coder.="-"; } else { $coder.="+"; }
   $output.=sprintf("%4s %2s %3s %6.1f %+4d%% %5.1f %4.1f %5.1f %5.1f %5.1f %6.1f %6.1f %5.1f %08d %5.1f\n",
    $w[1],$coder,substr($w[2],3,2),$w[17],int($w[34]),$w[13],$w[18],$w[3],$w[4],$w[5],$w[6],$w[22],$w[23],substr($w[31],6,2).substr($w[31],9,6),lukeutils::clock_wrap($w[11]+$stlocs->longitude($w[1])/360*24));
   if ($threshhold_flag) {$output=lukeutils::trim($output)." ".$w[35]."\n";}
   if ($label) {$output=lukeutils::trim($output)." $label\n";}
  }
  return $output;
 }

#Returns whether negative errors are present indicating qm_flag results are here
#Arguments: none
####
sub Check_For_qmflag_Info
{
 my $self = shift;
 
 my @w;
 foreach(@$self) { @w = split(' ',$_); if ($w[18]<0) {return 1;} }  
 return 0;
}

#Returns the number of points present, used, and rejected for the given tec_dump
#Arguments: none
####
sub Count_Points
{
    my $self = shift;
    
    my @w;
    my ($used, $rejected)=(0,0);
    foreach (@$self) { @w = split(' ',$_); if ($w[18]<0) { $rejected++; } else { $used++; } }
    return ($used+$rejected,$used,$rejected);
}


#Return statistics for each station in tec_dump
#Arguments: (column to analyze, label for max/min, whether or not to use qmflagged data)
#####
sub station_column_statistics
 {
  my ($self, $col, $label, $dataset) = @_;

  if (! $dataset) {$dataset = "both";}
  my ($good, $bad);
   if ($dataset eq "both" ) { $good=1; $bad=1;}
elsif ($dataset eq "used" ) { $good=1; $bad=0;}
else                        { $good=0; $bad=1;}
  my @worker;
  my %stats;
  my @order;

  foreach (@$self)
   {
    @worker=split(' ',$_);
    if (($worker[18]< 0) && (! $bad )) { next; }
    if (($worker[18]>=0) && (! $good)) { next; }
    if (!defined $stats{$worker[1]}) { $stats{$worker[1]}=new Statistics; @order=(@order,$worker[1]);}
    $stats{$worker[1]}->add($worker[$col-1],$worker[2]);
   }

  my $s;
  my ($n, $mean, $sigma, $min, $max, $rms, $minlabel, $maxlabel);

  $s.=sprintf("%4s  %6s  %7s  %7s  %7s  %7s  %5s  %7s  %5s\n"
             ,"site","npts","  mean ","sigma","rms","min"," sat","max"," sat");
  $s.=sprintf("%4s  %6s  %7s  %7s  %7s  %7s  %5s  %7s  %5s\n"
             ,"-"x4,"-"x6,"-"x7,"-"x7,"-"x7,"-"x7,"-"x5,"-"x7,"-"x5);

  foreach (@order)
   {
    ($n, $mean, $sigma, $min, $max, $rms, $minlabel, $maxlabel)=$stats{$_}->calc;
    $s.=sprintf("%4s  % 6d  % 7.2f  % 7.2f  % 7.2f  % 7.2f  %5s  % 7.2f  %5s\n"
               ,$_,$n,$mean,$sigma,$rms,$min,$minlabel,$max,$maxlabel);
    if ($label) {$s=lukeutils::trim($s)." $label\n";}
   }
  return $s;
 }

#Return statistics for each station in tec_dump
#Print additional lines for any statistics that depend on
#separation of rejected vs used data
#####
sub station_column_statistics_comparison
 {
  my ($self, $col, $label, $label2) = @_;

  my @worker;
  my %stats;
  my %stats_used;
  my @order;
  my $temp;
  my $temp2;
  my $flag;
  my $rejected;

  my $i=0;

  foreach (@$self)
   {
    @worker=split(' ',$_);
    if ($worker[18] < 0) { $rejected = 1; $worker[18]= -$worker[18]; } else { $rejected = 0; } #Ensure recorded errors are positive
    if (!defined $stats     {$worker[1]}) { $stats     {$worker[1]}=new Statistics; @order=(@order,$worker[1]);}
    if (!defined $stats_used{$worker[1]}) { $stats_used{$worker[1]}=new Statistics;}
    $stats     {$worker[1]}->add($worker[$col-1],$worker[2]);
    if ($rejected) { next; }
    $stats_used{$worker[1]}->add($worker[$col-1],$worker[2]);
   }

  my $s;
  my ($n, $mean, $sigma, $min, $max, $rms, $minlabel, $maxlabel);

  $s.=sprintf("%4s  %6s  %7s  %7s  %7s  %7s  %5s  %7s  %5s\n"
             ,"site","npts","  mean ","sigma","rms","min"," sat","max"," sat");
  $s.=sprintf("%4s  %6s  %7s  %7s  %7s  %7s  %5s  %7s  %5s\n"
             ,"-"x4,"-"x6,"-"x7,"-"x7,"-"x7,"-"x7,"-"x5,"-"x7,"-"x5);

  foreach (@order)
   {
    ($n, $mean, $sigma, $min, $max, $rms, $minlabel, $maxlabel)=$stats{$_}->calc;
    $temp =sprintf("%4s  % 6d  % 7.2f  % 7.2f  % 7.2f  % 7.2f  %5s  % 7.2f  %5s\n"
               ,$_,$n,$mean,$sigma,$rms,$min,$minlabel,$max,$maxlabel);

    ($n, $mean, $sigma, $min, $max, $rms, $minlabel, $maxlabel)=$stats_used{$_}->calc;
    $temp2=sprintf("%4s  % 6d  % 7.2f  % 7.2f  % 7.2f  % 7.2f  %5s  % 7.2f  %5s\n"
               ,$_,$n,$mean,$sigma,$rms,$min,$minlabel,$max,$maxlabel);

    if ($temp ne $temp2) { $flag = 1; } else { $flag = 0; }

    if ($label ) {$temp =lukeutils::trim($temp )." $label \n";}
    if ($label2) {$temp2=lukeutils::trim($temp2)." $label2\n";}

    $s.=$temp;
    if ($flag) { $s.=$temp2; }
   }
  return $s;
 }

#Prints only the lines in array 2 that are not present identically in line 1
#Used to compare station statistics for rejected and all-point data
#####
sub different_lines
{
    my ($arr1, $arr2) = @_;
    
    foreach my $testline (@$arr2)
    {
	if (! grep /$testline/,@$arr1) { print $testline; }
    }
}


#Return min and max STECDIFF percentage for each station in tec_dump
#Returns a mini tec_dump array for further processing
#Arguments: none
#####
sub station_stecdiff_minmax
 {
  my $self = shift;
  my @worker;
  my %stats;
  my @order;

  foreach (@$self)
   {
    @worker=split(' ',$_);
    if (!defined $stats{$worker[1]}) { $stats{$worker[1]}=new Statistics; @order=(@order,$worker[1]);}
    $stats{$worker[1]}->add($worker[34],$_);
   }

  my ($n, $mean, $sigma, $min, $max, $rms, $minlabel, $maxlabel);
  my @return_array;
  foreach (@order)
   {
    ($n, $mean, $sigma, $min, $max, $rms, $minlabel, $maxlabel)=$stats{$_}->calc;
    @return_array=(@return_array,$minlabel,$maxlabel);
   }
  return \@return_array;
 }


1;
