#!/usr/bin/python
#
# DateTime.py -- Date & Time class based on Timetrans library.
#                Represents a Date/Time as seconds past J2000 epoch (see JDREF constant)
#                and provides various format conversions and date delta arithmetic.
#

# Implemented as many small methods following Timetrans library.
# Object state is saved as a reference to a double value containing
# seconds past epoch.
#

import os, sys, re, string, math, time, types
import pdb

#CONSTANTS
JDREF       = 2451545.0   #Julian date of J2000 reference epoch
SECDAY      = 86400.0     #seconds in a day

MONTHS = ( 'JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN',
           'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC' )
MONTH_STRS = { 'JAN' : '01', 'FEB' : '02', 'MAR' : '03',
               'APR' : '04', 'MAY' : '05', 'JUN' : '06',
               'JUL' : '07', 'AUG' : '08', 'SEP' : '09',
               'OCT' : '10', 'NOV' : '11', 'DEC' : '12' }

INTERVALS = ( 'decade', 'year', 'mon', 'day', 'hour', 'min', 'sec', 'frac' )

#Supportive Function

def echo   (str          ): sys.stdout.write(str + "\n")
def err    (str          ): sys.stderr.write(str + "\n")
def warn   (str          ): err("---WARNING: "+str)
def die    (str, status=1): err("***ERROR: "+str); sys.exit(status)

def flatten(lister): #flattens incoming (potentially) multidimensional lists into a single list recursively
    master_list=[] #start with nothing in list
    for x in lister:
        if type(x) in (tuple,list): master_list.extend(flatten(x))
        else                      : master_list.append(        x )
    return master_list

def mod_remainder(first,divideby): #return the mod and remainder for a division
    mod       = int(first/divideby)
    remainder = first-mod*divideby
    return (mod, remainder)

def ymd2doy(year=0,month=0,day=0): #find integer day of year from year month and day
    (year,month,day)=flatten((year,month,day))[0:3]
    year = year_4digit(year)
    return date2j(year,month,day)-date2j(year,1,0)

def wraptime(year=0, month=0, day=0, hour=0, minute=0, second=0): #ensure times wrap correctly (not for day boundaries)
    (year,month,day,hour,minute,second)=flatten((year,month,day,hour,minute,second))[0:6]
    while second  <  0:
        second=second+60
        minute=minute-1
    while second >= 60:
        second=second-60
        minute=minute+1
    while minute  <  0:
        minute=minute+60
        hour  =hour-1
    while minute >= 60:
        minute=minute-60
        hour  =hour+1
    while hour   <   0:
        hour  =hour+24
        day   =day-1
    while hour  >=  24:
        hour  =hour-24
        day   =day+1
    while month  <   1:
        month =month+12
        year  =year - 1
    while month >=  12:
        month =month-12
        year  =year + 1
    return (year,month,day,hour,minute,second)
        

def calsec(year=0,month=0,day=0,hour=0,minute=0,second=0): #convert from list to sec
        (year,month,day,hour,minute,second)=flatten((year,month,day,hour,minute,second))[0:6]
        year = year_4digit(year)
        jd   = date2j(year,month,day) - 0.5
        s    = jd2sec(jd) + hour*3600 + minute*60 + second
        return s

def year_4digit(year): #ensure 4 digit year
        if (year < 100): #2 digit year
             if (year>68): year = year + 1900
             else        : year = year + 2000
        return year

def seccal(seconds): #convert seconds past J2000 to calendar list
        frac=seconds-int(seconds)
        if (frac < 0): frac+=1
        (year,month,day,hour,minute,second) = jd2cal(sec2jd(seconds-frac+0.5))
        return (year,month,day,hour,minute,int((second+frac-0.5)*100)/100.0)

def sec2jd(seconds): #convert seconds past J2000 to Julian date (days)
        return seconds / SECDAY + JDREF

def jd2cal(juliandate): #convert julian date to calendar list
        jdplus = juliandate + 0.5
        jdint = int(jdplus)
        dsec = SECDAY*(jdplus - int(jdplus))
        if (dsec >= SECDAY):
            jdint = jdint + 1
            dsec  = dsec  - SECDAY
        isec = int(dsec)
        frac = dsec-isec
        (year,month,day)=j2date(jdint)
        hour   = int(isec / 3600)
        isec   = isec - hour * 3600
        minute = int(isec / 60  )
        isec   = isec - minute * 60
        second = isec + frac
        return (year,month,day,hour,minute,second)

def date2j(year,month,day): #convert YYYYMMDD list to integer Julian date
        (year,month,day)=flatten((year,month,day))[0:3]
        if (month > 2):
            month = month - 3
        else:
            month = month + 9
            year = year - 1
        c = int(year / 100)
        ya = year - c*100
        jd = int((146097*c)/4) + int((1461*ya)/4)+int((153*month+2)/5) + day + 1721119
        return jd

def j2date(jd): #convert integer Julian date to YYYYMMDD list
        j = int(jd)
        j = j - 1721119
        y = int((4*j-1)/146097)
        j=4*j-1-146097*y
        d=int(j/4)
        j=int((4*d+3)/1461)
        d=4*d+3-1461*j
        d=int((d+4)/4)
        m=int((5*d-3)/153)
        d=5*d-3-153*m
        d=int((d+5)/5)
        y=100*y+j
        if (m<10):
            m=m+3
        else:
            m=m-9
            y=y+1
        return (y,m,d)

def jd2sec(jd): #convert double Julian date to seconds J2000
        return (jd - JDREF) * SECDAY

def ch2cal(calstring): #convert calendar string dd-mmm-yyyy hh:mm:ss.frac to calendar list
        calstringpattern=re.compile("([0-9]{1,2})\-([a-zA-Z]{3})\-([0-9]{4}).?([0-9]{0,2}):?([0-9]{0,2}):?([0-9]{0,2}\.?[0-9]*)")
        calstringmatch=calstringpattern.match(calstring)
        if calstringmatch:
            (sday,smonth,syear,shour,sminute,ssecond)=calstringmatch.groups()
            smonth=smonth.upper()
            if (smonth in MONTH_STRS.keys()):
                day   = float(sday)
                year  = float(syear)
                month = float(MONTH_STRS[smonth])
                hour  = float(shour)
                minute = float(sminute)
                second = float(ssecond)
                return(year,month,day,hour,minute,second)
            else:
                die("Illformed month in CalStr: "+smonth+" from calstring "+calstring)
        else:
            die("DateTimeClass: Illformed CalStr: "+calstring)

def ch2sec(calstring): #convert calendar string dd-mmm-yyyy hh:mm:ss.frac to sec J2000
        return calsec(ch2cal(calstring))

def delta2dsec(deltastring): #convert delta time [+-]dddd/hh:mm:ss.frac to delta seconds
        sign=1
        deltapattern=re.compile("([-+]?)([0-9]{0,4})/?([0-9]{0,2}):?([0-9]{0,2}):?([0-9]{0,2}\.?[0-9]*)")
        deltamatch=deltapattern.match(deltastring)
        if (deltamatch):
            (sign_str,sday,shour,sminute,ssecond) = deltamatch.groups()
            if (sign_str == "-"): sign = -1
            day = float(sday)
            hour = float(shour)
            minute = float(sminute)
            second = float(ssecond)
            ds = day*86400 + hour*3600 + minute*60 + second
            ds = ds * sign
            return ds
        else:
            die ("DateTimeClass: Illformed Delta Time String : "+deltastring)

def dsec2delta(deltaseconds): #convert delta seconds to 5 delta time values (sign, day, hour, minute, second)
        if (ds >= 0): sign = "+"
        else        : sign = "-"
        (day   ,ds) = mod_remainder(ds,86400)
        (hour  ,ds) = mod_remainder(ds,3600 )
        (minute,ds) = mod_remainder(ds,60   )
        second     = ds
        return (sign,day,hour,minute,second) 

def dsec2delta_str(deltaseconds): #convert delta seconds to Delta Time String [+-]dddd/hh:mm:ss.frac
        (sign,day,hour,minute,second)=sec2delta(deltaseconds)
        return "%s%d/%02d:%02d:%06.3f"%(sign,day,hour,minute,second)

def calstring2callist(calstring): #convert calstring yyyy/mm/dd,hh:mm:ss.frac to callist
        calstringpattern=re.compile("([0-9]{2,4})/([0-9]{2})/([0-9]{2}).?([0-9]{0,2}):?([0-9]{0,2}):?([0-9]{0,2}\.?[0-9]*)")
        calmatch = calstringpattern.match(calstring)
        if not calmatch: die ("DateTimeClass: Illformed calstring: "+calstring)
        (year,month,day,hour,minute,second) = calmatch.groups()
        year=float(year)
        if year > 99 and year < 1900 : die("DateTimeClass: Illformed calstring (bad year): "+calstring)
        return (year,float(month),float(day),float(hour),float(minute),float(second))

#Definition Class
class DateTimeClass:
    "Converts dates and times between a variety of formats"

    def __init__(self,init=0.0):
            self.seconds = 0
            if (type(init)==     int): #integer is SJ2000
                self.seconds = init
            if (type(init)==   float): #float is Sj2000
                self.seconds = init
            if (isinstance(init,DateTimeClass)): #another DateTime class was passed
                self.seconds = init.toJ2000()
            if (type(init)==str     ): #Calstring
                self.fromCalString(init)

    def copy(self): #make a copy of this object, don't return the same object
        return DateTimeClass().fromJ2000(self.toJ2000())

    def fromCalList(self,year=0,month=0,day=0,hour=0,minute=0,second=0): #set object from a Calendar List
        #Gotta handle people sending a single list (y,m,d,h,m,s) or each separately y,m,d,h,m,s
        self.seconds = calsec(flatten((year,month,day,hour,minute,second))[0:6])
        return self

    def fromNow(self): #take the date/time from today, right now
        (year,month,day,hour,minute,second)=time.localtime()[0:6]
        self.fromCalList(year,month,day,hour,minute,second)
        return self

    def fromNowUTC(self): #take the date/time from today, right now, in utc
        (year,month,day,hour,minute,second)=time.gmtime()[0:6]
        self.fromCalList(year,month,day,hour,minute,second)
        return self

    def toCalList(self): #write object as a Calendar List
        return seccal(self.seconds)

    def fromCalString(self,calstring): #set object from a Calendar String yyyy/mm/dd,hh:mm:ss.frac
        self.fromCalList(calstring2callist(calstring))
        return self
    
    def toCalString(self): #write object as a Calendar string yyyy/mm/dd,hh:mm:ss.frac
        return "%4.4d/%02d/%02d,%02d:%02d:%06.3f"%self.toCalList()

    def fromYYMMDD(self,yymmdd): #set object from a YYMMDD string
        self.fromCalList(year_4digit(float(yymmdd[0:2])),float(yymmdd[2:4]),float(yymmdd[4:6]))
        return self

    def fromYYYYMMDD(self,yymmdd): #set object from a YYYYMMDD string
        self.fromCalList(float(yymmdd[0:4]),float(yymmdd[4:6]),float(yymmdd[6:8]))
        return self

    def fromYYMMDDHHMMSS(self,ymdhms): #set object from a YYMMDDHHMMSS string
        self.fromCalList(year_4digit(float(ymdhms[0:2])),float(ymdhms[2:4]),float(ymdhms[4:6]),
                         float(ymdhms[6:8]),float(ymdhms[8:10]),float(ymdhms[10:12]))
        return self

    def fromYYYYMMDDHHMMSS(self,ymdhms): #set object from a YYYYMMDDHHMMSS string
        self.fromCalList(float(ymdhms[0:4]),float(ymdhms[4:6]),float(ymdhms[6:8]),
                         float(ymdhms[8:10]),float(ymdhms[10:12]),float(ymdhms[12:14]))
        return self

    def toYYMMDDHHMMSS(self): #output a YYMMDDHHMMSS string
        (year,month,day,hour,minute,second)=self.toCalList()
        return "%s%02d%02d%02d%02d%02d"%(("%04d"%(year))[2:4],month,day,hour,minute,second)

    def toHHMMSS(self): #output a HHMMSS string
        (year,month,day,hour,minute,second)=self.toCalList()
        return "%02d%02d%02d"%(hour,minute,second)

    def toYYYYMMDDHHMMSS(self): #output a YYYYMMDDHHMMSS string
        (year,month,day,hour,minute,second)=self.toCalList()
        return "%04d%02d%02d%02d%02d%02d"%(year,month,day,hour,minute,second)

    def toYYYYMMDD(self): #write object as YYYYMMDD string
        (year,month,day,hour,minute,second)=self.toCalList()
        return "%04d%02d%02d"%(year,month,day)

    def toYYYY(self): #write object as just 4 digit year
        (year,month,day,hour,minute,second)=self.toCalList()
        return "%04d"%(year)

    def toYYMMDD(self): #write object as YYMMDD string
        (year,month,day,hour,minute,second)=self.toCalList()
        return "%s%02d%02d"%(("%04d"%(year))[2:4],month,day)

    def toHoursInDay(self): #write object as Hours In Day (fraction)
        (year,month,day,hour,minute,second)=self.toCalList()
        return hour+minute/60.0+second/60.0/60.0

    def toSecondsInDay(self): #write object as seconds in day (fraction)
        (year,month,day,hour,minute,second)=self.toCalList()
        return hour*60*60+minute*60+second

    def toJ2000(self): #write object as seconds past J2000
        return self.seconds

    def fromJ2000(self,second=0): #set object from seconds past J2000
        self.seconds=second
        return self

    def toSecPastCalstring(self,calstring): #write object as delta seconds since a specified date/time
        return DateTimeClass(calstring).toJ2000()-self.seconds

    def toSecPastSeconds(self,seconds): #write object as delta seconds since a specified J2000 seconds
        return seconds - self.seconds

    def toSecPastCalList(self,year=0,month=0,day=0,hour=0,minute=0,second=0): #write object as delta seconds since a specified CalList
        return DateTimeClass().fromCalList(year,month,day,hour,minute,second).toJ2000() - self.seconds

    def toDOY(self): #write object as Day of Year
        (year,month,day,hour,minute,second)=self.toCalList()
        return ymd2doy(year,month,day)

    def fromCh(self,chstring): #set object from char string dd-mon-yyyy hh:mm:ss.frac
        self.fromCalList(ch2cal(chstring))
        return self

    def toCh(self): #write object as char string dd-mon-yyyy hh:mm:ss.frac
        (year,month,day,hour,minute,second)=self.toCalList()
        smonth=MONTHS[month-1]
        return "%02d-%3s-%04d %02d:%02d:%06.3f"%(day,smonth,year,hour,minute,second)

    def zero_out_hhmmss(self): #make it midnight of the current day
        (year,month,day,hour,minute,second)=self.toCalList()
        self.fromCalList(year,month,day,0,0,0)
        return self

    def toLocalTime(self,longitude): #figure out current local time (hours) at given longitude
        localtime = self.toHoursInDay() + longitude/360*24
        while (localtime <  0): localtime=localtime+24
        while (localtime > 24): localtime=localtime-24
        return localtime

    def floor(self,interval): #round current object to a specified accuracy
        (year,month,day,hour,minute,second)=self.toCalList()
        interval=interval.lower()
        if   (interval.find('year')): self.fromCalList(year,        0,   0,     0,          0)
        elif (interval.find('mon' )): self.fromCalList(year,month,  0,   0,     0,          0)
        elif (interval.find('day' )): self.fromCalList(year,month,day,   0,     0,          0)
        elif (interval.find('hour')): self.fromCalList(year,month,day,hour,     0,          0)
        elif (interval.find('min' )): self.fromCalList(year,month,day,hour,minute,          0)
        elif (interval.find('sec' )): self.fromCalList(year,month,day,hour,minute,int(second))
        else                        : die("DateTimeClass: Floor: Malformed interval: "+interval)
        return self

    def AddDeltaCalList(self,year=0,month=0,day=0,hour=0,minute=0,second=0):
        (year,month,day,hour,minute,second)=flatten((year,month,day,hour,minute,second))[0:6]
        (cyear,cmonth,cday,chour,cminute,csecond)=self.toCalList()
        self.fromCalList(cyear+year,cmonth+month,cday+day,chour+hour,cminute+minute,csecond+second)
        return self

    def AddDeltaCalString(self,calstring):
        ( year, month, day, hour, minute, second)=calstring2callist(calstring)
        (cyear,cmonth,cday,chour,cminute,csecond)=   self.toCalList()
        self.fromCalList(cyear+year,cmonth+month,cday+day,chour+hour,cminute+minute,csecond+second)
        return self

    def AddDeltaSeconds(self,seconds):
        self.seconds=self.seconds+seconds
        return self

    def AddDeltaDays(self,days):
        self.seconds=self.seconds+days*24*60*60
        return self


def main():
    print "Testing DateTime.py implementation... hold onto your hats..."
#    pdb.set_trace()

    print "Testing StrCal -> StrCal"
    start="2002/02/02,02:02:02.02"
    dater = DateTimeClass(start)
    print start,"was understood as",dater.toCalString()
    if (not dater.toCalString() == "2002/02/02,02:02:02.020"): die ("***CalString From->To FAILED...")

    sample = 65887322
    dater.fromJ2000(sample)
    print sample,"J2000 understood as",dater.toJ2000()
    if (not dater.toJ2000() == sample): die ("***J2000 From->To FAILED...")

    start2="05-APR-2001 01:01:01.011"
    dater.fromCh(start2)
    print start2
    print dater.toCh()
    print start2,"was understood as",dater.toCh()
    if not dater.toCh() == start2: die ("***Ch From->To FAILED...")

    print "DOY calculated as",dater.toDOY()
    if (not dater.toDOY() == 95): die ("***toDOY FAILED...")

    print "Floor testing, rounding to day accuracy yields ",dater.floor('day').toCalString()

#    print "Testing for 60 second problem"
#    print DateTimeClass().fromJ2000(139754880.0).toCalString()

if __name__ == "__main__":
  main()
