#
# Stats.pm -- Simple statistics class:  computes mean, sigma, min, max, rms.
#
# Author: Brian Wilson
#    @(#) Stats.py     1.0     2003/11/24
#
# Implemented by saving five accumulators:
#   no of points, mean, sum of squares of diffs from mean, min, and max.
# Methods:
#   add    -- add a data point to the accumulating stats
#   calc   -- compute the five statistics:  n, mean, std dev, min, max, rms
#   label  -- set the label for printing
#   __repr__   -- generates one-line string version of statistics for easy printing
#   reset  -- zero the accumulators
#   addm   -- add an array of data points to the accumulators (add multiple)
#
# See tests at end of file for example usage.
#

from math import sqrt

class Stats:
    def __init__(self):
        self.reset()

    def add(self, val):
        self.n += 1
        dval = val - self.mean     # difference from current mean
        self.mean += dval/self.n   # update mean
        dval = val - self.mean     # diff from new mean
        self.sumsq += dval*dval    # update sum of squares
        if ( self.n == 1 ):
            self.min = val
            self.max = val
        else:
            self.min = min(self.min, val)
            self.max = max(self.max, val)
        return self

    def calc(self):
        sigma = 0.; rms = 0.
        if (self.n > 0):
            if (self.n >= 2):
                sd2 = self.sumsq / (self.n-1)
                if (sd2 > 0.): sigma = sqrt(sd2)
                else: sigma = 0.
            rms = sqrt(self.mean*self.mean +  self.sumsq/self.n)
        return (self.n, self.mean, sigma, self.min, self.max, rms)

    def label(self, str):
        self.labelStr = str
        return self
        
    def __repr__(self):
        if (self.labelStr == None or self.labelStr == ""): self.labelStr = "Stats"
        line = self.labelStr + ": "
        line += "%d %f %f %f %f %f" % self.calc()
        return line
#        out.write(line)

    def reset(self):
        self.n = 0
        self.mean = 0.0
        self.sumsq = 0.0
        self.min = 0.0
        self.max = 0.0
        self.labelStr = ""
        return self

    def addm(self, seq):
        for val in seq:
            self.add(val)
        return self


if __name__ == '__main__':
    import sys
    def echo2(str): sys.stderr.write(str + "\n")
    def die(str, status=1): echo2(str); sys.exit(status)

    vals = ( 2.3, 4.5, 1.8, 6.2, 3.5 )
    print "\nStats:  start testing . . ."
    print "test1:  Should get zeros for a new Stats object, printed without a label."
    s = Stats()
    print s

    print "\ntest2:  Add points one by one and print stats with label."
    print "        Adding ( 2.3, 4.5, 1.8, 6.2, 3.5 )."
    for v in vals:
        s.add(v)
    print "Columns:  N  Mean  StdDev  Min  Max  RMS"
    s.label("test2")
    print s

    print "\ntest3:  Reset accumulators and print what should be zeros."
    s.reset()
    print s.label("test3")

    print "\ntest4:  Add an array of points (using addm) and print stats;"
    print "        should get same answer as test2 above."
    s = Stats().label("test4").addm(vals)
    print s

    print "\ntest5:  Call addm with bad arg to show error message."
    try:
        s.addm(5.4)
    except TypeError:
        print "test5: Got expected type error."
        echo2("Stats: Error, single argument to addm method must be of sequence type.")

    print "\nStats:  done testing.\n"
