#!/usr/local/bin/python
#
# coordinates.py -- Conversions between various spatial coordinate systems on a
#                   globe and handy spatial calculations therein. 
#                   Also includes some wonderful new smart functions that perform
#                   desired transformations on a Do The Right Thing basis
#

import types
from math import *
from ionoutil import *
import numpy as N

#import pdb

##CONSTANTS

D2R = pi / 180.0
R2D = 180.0 / pi
EARTH_FLATTENING_INVERSE = 298.257223563   # no units
EARTH_SEMI_MAJOR_AXIS    = 6378137         # m
EARTH_SEMI_MINOR_AXIS    = 6356752         # m
EARTH_RADIUS             = (EARTH_SEMI_MAJOR_AXIS+EARTH_SEMI_MINOR_AXIS)/2.0 
NORTH_MAGNETIC_POLE_LATITUDE  =  78.982    # deg
NORTH_MAGNETIC_POLE_LONGITUDE = -70.905    # deg
DEFAULT_SHELL_HEIGHT          = 450e3      # m

THETA_MAG   = 0.5*pi-D2R*NORTH_MAGNETIC_POLE_LATITUDE  #radian
PHI_MAG     = D2R*NORTH_MAGNETIC_POLE_LONGITUDE-1.5*pi #radian

#TRANSFORMATIONS FOR MAGNETIC COORDINATES
T11     =  cos(THETA_MAG)*sin(PHI_MAG)
T12     = -cos(THETA_MAG)*cos(PHI_MAG)
T13     = -sin(THETA_MAG)
T21     =  cos(PHI_MAG)
T22     =  sin(PHI_MAG)
T23     =  0
T31     =  sin(THETA_MAG)*sin(PHI_MAG)
T32     = -sin(THETA_MAG)*cos(PHI_MAG)
T33     =  cos(THETA_MAG)

# DEFINE terms
# All rll are (meters,degrees,degrees)
# All xyz are (meters,meters,meters)
#
# xyzgeo     Earth centered earth fixed cartesian with Z defined along spin axis
# xyzmag     Earth centered             cartesian with Z defined along magnetic axis (dipole approximation)
# rllgeo     Earth centered earth fixed spherical (standard map coords) which does not distinguish a surface of the earth
# rllmag                                spherical with lat=90 defined along magnetic axis (dipole approxmiation)
# allgeo                                sperhical using WS84 ref ellipsoid as earth's surface (hence alt, not rad) with lat=90 as spin axis
# allmag                                spherical using WS84 ref ellipsoid as earth's surface but measure lat=90 as magnetic axis

# BASE TRANSFORMATIONS
# We'll use rllgeo as central conversions


#Take the elevation (in degrees) and return the mapping function
def mapping_function(el):
    RADIUS_EARTH             = 6370.0 #km
    IONOSPHERIC_SHELL_HEIGHT = 450.0  #km
    PI                       = 3.141592654
    t=cos(el/180.0*PI)/(1+IONOSPHERIC_SHELL_HEIGHT/RADIUS_EARTH)
    return 1.0/sqrt(1.0-t*t)

def xyzgeo2rllgeo(xyz):
   (x,y,z)=xyz
   lat    =R2D*atan2(z,sqrt(x*x + y*y))
   lon    =R2D*atan2(y,x              )
   rad    =sqrt (x*x + y*y + z*z)
   return (rad,lat,lon)

def rllgeo2xyzgeo(rll):
   (rad,lat,lon)=rll
   lat    =D2R*lat
   lon    =D2R*lon
   c      =rad*cos(lat)
   x      =c*cos(lon)
   y      =c*sin(lon)
   z      =rad*sin(lat)
   return (x,y,z)

def xyzgeo2allgeo(xyz):
   (x,y,z)=xyz
#   pdb.set_trace()
   f   = 1.0 / EARTH_FLATTENING_INVERSE
   e2  = ( 2.0 - f ) * f   
   e2p = e2 / ( 1.0 - e2 )
   re  = EARTH_SEMI_MAJOR_AXIS
   rp  = EARTH_SEMI_MINOR_AXIS
    
   d   = sqrt (x*x  +  y*y)
   lat = atan2(z*re , d*rp)
   sn  = sin(lat)
   cs  = cos(lat)

   lon = atan2(y , x)
   lat = atan2(z + e2p*rp*sn*sn*sn , abs(d - e2*re*cs*cs*cs))

   sn  = sin(lat)
   cs  = cos(lat)
   c   = 1.0/sqrt(re*re*cs*cs + rp*rp*sn*sn)

   if abs(lat) > 0.25*pi:
       alt = z/sn - rp*rp*c
   else:
       alt = d/cs - re*re*c

   return (alt, lat*R2D, lon*R2D)

def allgeo2xyzgeo(all):
   (alt, lat, lon)=all

   lat*=D2R
   lon*=D2R
   sn = sin(lat)
   cs = cos(lat)
   a2 = EARTH_SEMI_MAJOR_AXIS * EARTH_SEMI_MAJOR_AXIS
   b2 = EARTH_SEMI_MINOR_AXIS * EARTH_SEMI_MINOR_AXIS
   c  = 1.0/sqrt(a2*cs*cs + b2*sn*sn)
   x  = (a2*c + alt)*cs*cos(lon)
   y  = (a2*c + alt)*cs*sin(lon)
   z  = (b2*c + alt)*sn
   return (x, y, z)

def xyzgeo2xyzmag(xyz):
   (x, y, z)=xyz

   x2 = T11*x + T12*y + T13*z
   y2 = T21*x + T22*y + T23*z
   z2 = T31*x + T32*y + T33*z
   return (x2, y2, z2)

def xyzmag2xyzgeo(xyz):
   (x, y, z)=xyz

   x2 = T11*x + T21*y + T31*z
   y2 = T12*x + T22*y + T32*z
   z2 = T13*x + T23*y + T33*z
   return (x2, y2, z2)

#Find the intersection of a line with a sphere, line specified between p1 and p2
def line_to_sphere_intersection(p1, p2, R):
   (x1, y1, z1) = p1
   (x2, y2, z2) = p2
   (dx, dy, dz) = (x2 - x1,y2 - y1,z2 - z1)

   #calculate IPP point... there are two intersections however
   if abs(dz) == max(abs(dy),abs(dz)) and not dz == 0.0: #parameterize in terms of z, dz must not be zero
         a1 = dx/dz
         b1 = dy/dz
         a2 = x1 - a1*z1
         b2 = y1 - b1*z1
         ins=4.0*(a1*a2 + b1*b2)**2 - 4.0*(1 + a1**2 + b1**2)*(a2**2 + b2**2 - R**2)
         if ins < 0: return (None,None) #no intersection
         zipp1 = -(a1*a2 + b1*b2 + 0.5*sqrt(ins))/(1.0 + a1**2 + b1**2)
         zipp2 = -(a1*a2 + b1*b2 - 0.5*sqrt(ins))/(1.0 + a1**2 + b1**2)
         xipp1 = a1*zipp1 + a2
         yipp1 = b1*zipp1 + b2
         xipp2 = a1*zipp2 + a2
         yipp2 = b1*zipp2 + b2
   elif not dy == 0.0: #parameterize in terms of y, dy must not be zero
         a1 = dx/dy
         b1 = dz/dy
         a2 = x1 - a1*y1
         b2 = z1 - b1*y1
         ins=4.0*(a1*a2 + b1*b2)**2 - 4.0*(1 + a1**2 + b1**2)*(a2**2 + b2**2 - R**2)
         if ins < 0: return (None,None) #no intersection
         yipp1 = -(a1*a2 + b1*b2 + 0.5*sqrt(ins))/(1.0 + a1**2 + b1**2)
         yipp2 = -(a1*a2 + b1*b2 - 0.5*sqrt(ins))/(1.0 + a1**2 + b1**2)
         xipp1 = a1*yipp1 + a2
         zipp1 = b1*yipp1 + b2
         xipp2 = a1*yipp2 + a2
         zipp2 = b1*yipp2 + b2
   else: #everything but dx is zero? Solution lies in a straight line
         ins=R*R - y1*y1 - z1*z1
         if ins < 0: return (None, None)
         xipp1 =  sqrt(ins)
         yipp1 = y1
         zipp1 = z1
         xipp2 = -sqrt(ins)
         yipp2 = y1
         zipp2 = z1

   return (Coordinate((xipp1, yipp1, zipp1)), Coordinate((xipp2, yipp2, zipp2)))

###DERIVED TRANSFORMATIONS
def rllgeo2allgeo(rll): return xyzgeo2allgeo(rllgeo2xyzgeo(rll))
def allgeo2rllgeo(all): return xyzgeo2rllgeo(allgeo2xyzgeo(all))
def rllgeo2rllmag(rll): return xyzgeo2rllgeo(xyzgeo2xyzmag(rllgeo2xyzgeo(rll)))
def rllmag2rllgeo(rll): return xyzgeo2rllgeo(xyzmag2xyzgeo(rllgeo2xyzgeo(rll)))
def xyzmag2rllgeo(xyz): return xyzgeo2rllgeo(xgzmag2xyzgeo(xyz))
def rllgeo2xyzmag(rll): return xyzgeo2xyzmag(rllgeo2xyzgeo(rll))
def xyzgeo2rllmag(xyz): return rllgeo2rllmag(xyzgeo2rllgeo(xyz))
def xyzmag2rllmag(xyz): return xyzgeo2rllgeo(xyz)
def rllmag2xyzmag(rll): return rllgeo2xyzgeo(rll)
def xyzmag2allgeo(xyz): return xyzgeo2allgeo(xyzmag2xyzgeo(xyz))
def allgeo2xyzmag(all): return xyzgeo2xyzmag(allgeo2xyzgeo(all))
def rllmag2allgeo(rll): return rllgeo2allgeo(rllmag2rllgeo(rll))
def allgeo2rllmag(all): return rllgeo2rllmag(allgeo2rllgeo(all))

def ensure_num(lister):
   if not isinstance(lister[0],types.FloatType):
      return tuple([float(s) for s in lister])
   return lister

class Coordinate:
   "Handles converssions between coordinate systems of xyz/rll/all and geo/mag, internal representation is xyzgeo"
   def __init__(self,input=None,format=None):
      self.xyzgeo=(0,0,0)
      self.set(input,format)
   def set(self,input=None,format=None):
      if not input: return self
      if isinstance(input,Coordinate): self.xyzgeo = input.xyzgeo; return self
      if not format                  : self.xyzgeo = ensure_num(input) ; return self
      format=format.lower()
      if     format == "xyzgeo": self.xyzgeo =               ensure_num(input)
      elif   format == "xyzmag": self.xyzgeo = xyzmag2xyzgeo(ensure_num(input))
      elif   format == "rllgeo": self.xyzgeo = rllgeo2xyzgeo(ensure_num(input))
      elif   format == "rllmag": self.xyzgeo = rllmag2xyzgeo(ensure_num(input))
      elif   format == "allgeo": self.xyzgeo = allgeo2xyzgeo(ensure_num(input))
      else                     : die("Coordinates.py: unknown format: "+format)
      return self
   def to(self,format=None):
      if not format: return self.xyzgeo
      format=format.lower()      
      if     format == "xyzgeo": return self.xyzgeo
      if     format == "xyzmag": return xyzgeo2xyzmag(self.xyzgeo)
      if     format == "rllgeo": return xyzgeo2rllgeo(self.xyzgeo)
      if     format == "rllmag": return xyzgeo2rllmag(self.xyzgeo)
      if     format == "allgeo": return xyzgeo2allgeo(self.xyzgeo)
      die("Coordinates.py: unknown format: "+format)
   def radius(self): return sqrt(self.xyzgeo[0]**2+self.xyzgeo[1]**2+self.xyzgeo[2]**2)
   def __sub__(self,other): #subtraction is used to calculate distances
      if   isinstance(other,types.TupleType ) or isinstance(other,types.ListType): o=Coordinate(other)
      elif isinstance(other,Coordinate)                              : o=other
      else                                                           : die("Coordinate: unspecified subtraction type %s"%(type(other)))
      return sqrt((o.xyzgeo[0]-self.xyzgeo[0])**2+(o.xyzgeo[1]-self.xyzgeo[1])**2+(o.xyzgeo[2]-self.xyzgeo[2])**2)
   def __abs__(self): #return distance from origin(radius)
      return self.radius()
   def __cmp__(self,other): #compares distances as identical up to 1 mm
      if   isinstance(other,types.TupleType ) or isinstance(other,types.ListType): o=Coordinate(other)
      elif isinstance(other,Coordinate)                              : o=other
      else                                                           : die("Coordinate: unspecified comparison type %s"%(type(other)))
      return compareto(self.xyzgeo,o.xyzgeo,-2) 
   def __coerce__(self,other):
      if isinstance(other,types.FloatType ) or isinstance(other,types.IntType) or isinstance(other,types.LongType): return self.radius()
      if isinstance(other,types.StringType)                                                                       : return "(%f,%f,%f)"%self.xyzgeo
      if isinstance(other,types.ListType  ) or isinstance(other,types.TupleType)                                  : return self.xyzgeo
   def copy (self):
      n=Coordinate(self.xyzgeo)
      return n
   def localtime(self,time): #output the current local time at this location
      T=IonoTime(time)
      rll=xyzgeo2rllgeo(self.xyzgeo)
      return (rll[2]/360.0*24.0+T.to("HOURSINDAY"))%24
   def tolocaltime(self,time,desiredlocalhours): #rotate the current position until it aligns with the requested local time
      current_lt    =self.localtime(time)
      current_rllgeo=list(self.to("rllgeo"))
      current_rllgeo[2]+=(desiredlocalhours-current_lt)*360.0/24.0
      self.set(current_rllgeo,"rllgeo")
      return self
   def km2m(self): #convert current coordinates from km to m
      self.xyzgeo=(self.xyzgeo[0]*1e3,
                   self.xyzgeo[1]*1e3,
                   self.xyzgeo[2]*1e3)
      return self
   def m2km(self): #convert current coordinates from m to km
      self.xyzgeo=(self.xyzgeo[0]/1e3,
                   self.xyzgeo[1]/1e3,
                   self.xyzgeo[2]/1e3)
      return self
   def AzEl(self,ocord):
      #calculates AzEl from a pair of coords in degrees, standard is self is ground, ocord is satellite
      (x,y,z)=self.xyzgeo
      p=sqrt(x*x+y*y)
      if p == 0: die ("AzEl fails at poles")
      R=sqrt(x*x+y*y+z*z)
      e=N.zeros((3,3),"float32")
      e[0,0]= - y / p
      e[0,1]=   x / p
      e[0,2]= 0.0
      e[1,0]= - x*z / (p*R)
      e[1,1]= - y*z / (p*R)
      e[1,2]= p / R
      e[2,0]= x / R
      e[2,1]= y / R
      e[2,2]= z / R
      d=[0,0,0]
      for k in range(3):
         for i in range(3):
            d[k]+=(ocord.xyzgeo[i]-self.xyzgeo[i])*e[k,i]

      s = d[2] / sqrt(d[0]*d[0] + d[1]*d[1] + d[2]*d[2])
      if   s ==  1.0: El = 90
      elif s == -1.0: El = -90
      else          : El = R2D*N.arctan(s / sqrt(1.0 - s*s))

      if   (d[1] == 0.0) and (d[0] > 0.0): Az = 90
      elif (d[1] == 0.0) and (d[0] < 0.0): Az = 180+90
      elif (d[1] == 0.0)                 : Az = 0 #on top of each other
      else: 
         Az = R2D*N.arctan(d[0] / d[1])
         if   d[1] < 0.0: Az+=180
         elif d[1] > 0.0 and d[0] < 0.0: Az+=360
      return (Az,El)

   def project_raypath(self,Az,El,radius):
      #takes the existing XYZ location, an Azimuth, Elevation angle, and projects that vector out to some Shell Height to find the XYZ of that intercept

      PROJECTION_SLOP=100.0 #number of radii to project beyond to ensure intersection

      #get current lat,lon
      (r,lat,lon)=self.to("rllgeo")
      (x,y,z)    =self.xyzgeo

      #convert everything to radians
      lat*=D2R
      lon*=D2R
      Az*=D2R
      El*=D2R

      #figure out local direction information
      l_up    = sin(El)
      l_north = cos(El)*cos(Az)
      l_east  = cos(El)*sin(Az)

      sinlat= sin(lat)
      coslat= cos(lat)
      sinlon=-sin(lon)
      coslon= cos(lon)
    
      north_x = -sinlat * coslon
      north_y =  sinlat * sinlon
    
      up_x =  coslat * coslon
      up_y = -coslat * sinlon

      #figure out direction cosines
      dcos_x = l_up*up_x   + l_north*north_x  + l_east*sinlon
      dcos_y = l_up*up_y   + l_north*north_y  + l_east*coslon
      dcos_z = l_up*sinlat + l_north*coslat

      #project this unit vector from the current location out way past the radius of interest...
      distantpoint=Coordinate((x+dcos_x*radius*PROJECTION_SLOP,
                               y+dcos_y*radius*PROJECTION_SLOP,
                               z+dcos_z*radius*PROJECTION_SLOP))

      (p1,p2)=line_to_sphere_intersection(self.xyzgeo,distantpoint.xyzgeo,radius)
      if not p1 or not p2 : die("Coordinate: Project Raypath does not intersect sphere radius specified")


      #find the intersection closest to the distant location
      if p1-distantpoint < p2-distantpoint: return p1
      return p2

   def mapping_function(self,ocord):
      (Az,El)=self.AzEl(ocord)
      return mapping_function(El)

   def blocked_by_earth(self,ocoord):
      #figures out of the earth (simple radius calc) is blocking two satellites from seeing each other
      if not ocoord: return None
      (p1,p2)=line_to_sphere_intersection(self.xyzgeo,ocoord.xyzgeo,EARTH_RADIUS)
      return not (p1 == None and p2 == None)

   def tangent_point(self,ocoord):
      #calculates the tangent point for two reference points outside the ionosphere
      #failure to ensure points are outside generates warning, but extrapolation is performed regardless
      if not ocoord: return None
      x=N.array(self.xyzgeo  )
      y=N.array(ocoord.xyzgeo)
      #find normalized vector between self and ocoord
      xy=x-y
      xy/=sqrt(N.dot(xy,xy))
      z=y-N.dot(y,xy)*xy
      isbetween = N.dot(x-z,y-z) < 0 #figure out whether the tangent point is between the specified points or not
      return(Coordinate(tuple(z)),isbetween)
      
   def IPP(self,satloc=None,shell_height=DEFAULT_SHELL_HEIGHT):
      #calculates the ionospheric pierce point given a second (satellite) position and an optional shell height
      #returns a coordinate object
      if not satloc: return None
      R=EARTH_RADIUS+shell_height
      if not ( (satloc.radius() <= R and self.radius() >= R) or (satloc.radius() >= R and self.radius() <= R)): #one point within ionosphere shell, the other outside
         warn("Coordinate: IPP calculation called for points that do not span ionosphere... extrapolation will be attempted")
      (p1,p2)=line_to_sphere_intersection(self.xyzgeo,satloc.xyzgeo,R)
      if not p1 or not p2 : die("Coordinate: IPP calculation between points that do not intersect specified shell height")
      #find the intersection closest to both points
      d1=(p1-self) + (p1-satloc)
      d2=(p2-self) + (p2-satloc)
      if d1 < d2: return p1
      return p2

   def __repr__(self):
      (x,y,z)=self.xyzgeo
      return "(%f,%f,%f)"%(x,y,z)

###MAIN PROGRAM
def main():
   print "Checking base transformations to the 0.00000001 m accuracy"

   print "Testing xyzgeo2rllgeo",xyzgeo2rllgeo((50,40,30)) , (70.710678118654755, 25.104090250221386, 38.659808254090088)
   if not compareto(             xyzgeo2rllgeo((50,40,30)) , (70.710678118654755, 25.104090250221386, 38.659808254090088),-2) : die("FAILED xyzgeo2rllgeo test")
   print "Testing rllgeo2xyzgeo",rllgeo2xyzgeo((70.710678118654755, 25.104090250221386, 38.659808254090088)) , (50,40,30)
   if not compareto(             rllgeo2xyzgeo((70.710678118654755, 25.104090250221386, 38.659808254090088)) , (50,40,30),-2) : die("FAILED rllgeo2xyzgeo test")
   print "Testing xyzgeo2allgeo",xyzgeo2allgeo((5e6,4e6,3e6)) , (696793.828692437,25.237604550324996, 38.659808254090088)
   if not compareto(             xyzgeo2allgeo((5e6,4e6,3e6)), (696793.828692437, 25.237604550324996, 38.659808254090088),-2) : die("FAILED xyzgeo2allgeo test")
   print "Testing allgeo2xyzgeo",allgeo2xyzgeo((696793.828692437,25.237604550324996, 38.659808254090088)), (5e6,4e6,3e6)
   if not compareto(             allgeo2xyzgeo((696793.828692437,25.237604550324996, 38.659808254090088)), (5e6,4e6,3e6),2) : die("FAILED allgeo2xyzgeo test")

   print "Testing xyzgeo2xyzmag",xyzgeo2xyzmag((5000e3,4000e3,3000e3)), (-2678060.5653255386, 6033429.0600771112, 2534901.4547838513)
   if not compareto(             xyzgeo2xyzmag((5000e3,4000e3,3000e3)), (-2678060.5653255386, 6033429.0600771112, 2534901.4547838513),-2) : die("FAILED xyzgeo2xyzmag test")

   print "Testing xyzmag2xyzgeo",xyzmag2xyzgeo((-2678060.5653255386, 6033429.0600771112, 2534901.4547838513)),(5000e3,4000e3,3000e3)
   if not compareto(             xyzmag2xyzgeo((-2678060.5653255386, 6033429.0600771112, 2534901.4547838513)),(5000e3,4000e3,3000e3),-2) : die("FAILED xyzmag2xyzgeo test")

   print "Testing rllgeo2allgeo",rllgeo2allgeo((10e6,20,10)),(3624370.2502621757, 20.078914720833236, 10.0)
   if not compareto(             rllgeo2allgeo((10e6,20,10)),(3624370.2502621757, 20.078914720833236, 10.0),-2): die("FAILED rllgeo2allgeo test")

   print "Testing allgeo2rllgeo",allgeo2rllgeo((3624370.2502621757, 20.078914720833236, 10.0)),(10e6,20,10)
   if not compareto(             allgeo2rllgeo((3624370.2502621757, 20.078914720833236, 10.0)),(10e6,20,10), 0): die("FAILED allgeo2rllgeo test")

   print "All tests passed"
   
if __name__ == "__main__":
  main()
