#!/bin/env python

import sys, os
import scipy.io.mio as mio
from os.path import abspath
from plotlib import imageMap, marksOnMap
import matplotlib.pylab as M
import Numeric as N

USAGE = """
vtec_image.py [-o out.mpg] gaim_file.mat gaim_file2.mat . . .

If you are making a movie of multiple images, you should specify a fixed
colorbar scale with --vmin and --vmax.

If no output file is specified, the image or movie is displayed using an
interactive X window.

Optional switches:
 -h, --help: print USAGE
 -n, --nlevels: use <nlevels> contours/colors
 -o, --output:  specify output file
 -r, --range:   turn on RANGE mode: don't make images, print min/max VTEC for each file
 -t, --title:  specify title
     --xlabel: specify xaxis label
     --ylabel: specify yaxis label
     --vmin:   specify minimum value in colorbar
     --vmax:   specify maximum value in colorbar
     --imageWidth:   specify imageWidth  (default 1280)
     --imageHeight:  specify imageHeight (default 768)
 -p, --projection:   specify projection (default is moll for Mollweide, or use
                                         cyl for cylindrical/flat, robin for Robinson)

"""

def echo2(*s): sys.stderr.write(' '.join(s) + '\n')
def warn(*s): echo2('vtec_image:', *s)
def die(*s): warn('Error,', *s); echo2(USAGE); sys.exit()


def vtec_image(inFile, outFile, **options):
#def vtec_image(inFile, outFile, defaultNcontours=20,
#               defaultVmin=0., defaultVmax=40., **options):
    a = mio.loadmat( abspath(inFile) )
#    ncontours = options.get('ncontours', defaultNcontours)
#    vmin = options.get('vmin', defaultVmin)
#    vmax = options.get('vmax', defaultVmax)
#    if ncontours != defaultNcontours or vmax != defaultVmax:
#        delta = float(vmax) / ncontours
#        contourValues = [n*delta for n in range(ncontours+1)]
#    else:
#        contourValues = ncontours
    vmin = options.pop('vmin', None)
    vmax = options.pop('vmax', None)
    imageWidth = options.pop('imageWidth', 1280)
    imageHeight = options.pop('imageHeight', 768)
    projection = options.pop('projection', 'moll')
    map = imageMap(None, a['post_lon'], a['post_lat'], a['post_vtec'], vmin, vmax,
             imageWidth, imageHeight, outFile, projection, autoBorders=False, **options)
#    marksOnMap(map, lons, lats, projection='moll', colors='k')

def makeMovie(inFiles, outFile, options):
    if len(inFiles) == 1:
        inFile = inFiles[0]
        if 'title' not in options: options['title'] = 'VTEC Map from '+inFile 
        vtec_image(inFile, outFile, **options)
    else:
        if outFile:
            outMovieFile = os.path.splitext(outFile)[0] + '.mpg'
            files = []
        else:
            outf = None
        
        for inFile in inFiles:
            if outFile:
                outf = inFile + '.vtec.png'
                warn('Creating %s' % outf)
            options['title'] = 'VTEC Map from '+inFile
            vtec_image(inFile, outf, **options)
            M.clf()
            if outFile: files.append(outf)
        
        if outFile:
            cmd = 'convert ' + ' '.join(files) + ' ' + outMovieFile
            os.system(cmd)
            warn('Wrote %s' % outMovieFile)


def printRange(files, variable):
    for i, inFile in enumerate(files):
        a = mio.loadmat( abspath(inFile) )
        minr = N.minimum.reduce
        maxr = N.maximum.reduce
        vmin = minr(minr(a[variable]))
        vmax = maxr(maxr(a[variable]))
        print inFile, 'min/max: ', vmin, vmax
        if i == 0:
            gmin = vmin
            gmax = vmax
        else:
            gmin = min(gmin, vmin)
            gmax = max(gmax, vmax)
    print 'Global min/max: ', gmin, gmax

    
if __name__ == '__main__':
    from sys import argv
    import getopt
    me = argv[0]

    if len(argv) < 2: die('Too few arguments.')
    try:
        opts, argv = getopt.getopt(argv[1:], 'hn:o:p:rt:',
         ['help', 'nlevels=', 'output=', 'title=', 'xlabel=', 'ylabel=',
          'vmin=', 'vmax=', 'imageWidth=', 'imageHeight=', 'projection='])
    except getopt.GetoptError, (msg, bad_opt):
        die("%s error: Bad option: %s, %s" % (me, bad_opt, msg), USAGE)

    options = {}; outFile = None; rangeMode = False
    for opt, val in opts:
        if opt   in ('-h', '--help'):    echo2(USAGE); sys.exit()
        elif opt in ('-n', '--nlevels'): options['nlevels'] = int(val)
        elif opt in ('-o', '--output'):  outFile = val
        elif opt in ('-r', '--range'):   rangeMode = True
        elif opt in ('-t', '--title'):   options['title'] = val
        elif opt in ('--xlabel'):        options['xlabel'] = val
        elif opt in ('--ylabel'):        options['ylabel'] = val
        elif opt in ('--vmin'):          options['vmin'] = float(val)
        elif opt in ('--vmax'):          options['vmax'] = float(val)
        elif opt in ('--imageWidth'):    options['imageWidth'] = int(val)
        elif opt in ('--imageHeight'):   options['imageHeight'] = int(val)
        elif opt in ('-p', '--projection'):  options['projection'] = val

    if rangeMode:
        printRange(argv, 'post_vtec')
    else:
        makeMovie(argv, outFile, options)

